import torch
import torch.nn.functional as F

def buffered_arange(max):
    if not hasattr(buffered_arange, "buf"):
        buffered_arange.buf = torch.LongTensor()
    if max > buffered_arange.buf.numel():
        buffered_arange.buf.resize_(max)
        torch.arange(max, out=buffered_arange.buf)
    return buffered_arange.buf[:max]

def sample_word(logits, strategy='topp', **kwargs):
    if strategy == 'topp':
        return sample_top_p(logits, **kwargs)
    else:
        raise NotImplementedError

def sample_top_p(logits, k=5, sample=False, **kwargs):
    if sample:
        v, ix = torch.topk(logits, k)
        out = logits.clone()
        out[out < v[..., [-1]]] = -float("Inf")    
        probs = F.softmax(out, dim=-1)

        ix = torch.multinomial(probs.squeeze(0), num_samples=1)
    else:
        _, ix = torch.topk(logits, k=1, dim=-1)
    
    next_words = ix

    return next_words

def item(tensor):
    if hasattr(tensor, "item"):
        return tensor.item()
    if hasattr(tensor, "__getitem__"):
        return tensor[0]
    return tensor

def apply_to_sample(f, sample):
    if hasattr(sample, "__len__") and len(sample) == 0:
        return {}
    
    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, dict):
            return {key: _apply(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        elif isinstance(x, tuple):
            return tuple(_apply(x) for x in x)
        elif isinstance(x, set):
            return {_apply(x) for x in x}
        else:
            return x
    
    return _apply(sample)

def move_to_cuda(sample, device=None):
    device = device or torch.cuda.current_device()

    def _move_to_cuda(tensor):
        # non_blocking is ignored if tensor is not pinned, so we can always set
        # to True (see github.com/PyTorchLightning/pytorch-lightning/issues/620)
        return tensor.to(device=device, non_blocking=True)

    return apply_to_sample(_move_to_cuda, sample)

def move_to_cpu(sample):
    def _move_to_cpu(tensor):
        # PyTorch has poor support for half tensors (float16) on CPU.
        # Move any such tensors to float32
        if tensor.dtype in {torch.bfloat16, torch.float16}:
            tensor = tensor.to(dtype=torch.float32)
        return tensor.cpu()
    return apply_to_sample(_move_to_cpu, sample)


import logging
from contextlib import contextmanager

import torch
import torch.nn.functional as F

logger = logging.getLogger(__name__)

def item(tensor):
    if hasattr(tensor, "item"):
        return tensor.item()
    if hasattr(tensor, "__getitem__"):
        return tensor[0]
    return tensor

def get_rng_state():
    state = {"torch_rng_state": torch.get_rng_state()}
    if torch.cuda.is_available():
        state["cuda_rng_state"] = torch.cuda.get_rng_state()
    return state

def set_rng_state(state):
    torch.set_rng_state(state["torch_rng_state"])
    if torch.cuda.is_available():
        torch.cuda.set_rng_state(state["cuda_rng_state"])

class set_torch_seed(object):
    def __init__(self, seed):
        assert isinstance(seed, int)
        self.rng_state = get_rng_state()

        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        set_rng_state(self.rng_state)

def apply_to_sample(f, sample):
    if hasattr(sample, "__len__") and len(sample) == 0:
        return {}
    
    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, dict):
            return {key: _apply(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        elif isinstance(x, tuple):
            return tuple(_apply(x) for x in x)
        elif isinstance(x, set):
            return {_apply(x) for x in x}
        else:
            return x
    
    return _apply(sample)

def move_to_cuda(sample, device=None):
    device = device or torch.cuda.current_device()

    def _move_to_cuda(tensor):
        return tensor.to(device=device, non_blocking=True)
    
    return apply_to_sample(_move_to_cuda, sample)

def move_to_cpu(sample):
    def _move_to_cpu(tensor):
        if tensor.dtype in {torch.bfloat16, torch.float16}:
            tensor = tensor.to(dtype=torch.float32)
        return tensor.cpu()
    return apply_to_sample(_move_to_cpu, sample)

def prepare_sample(sample):
    if torch.cuda.is_available():
        sample = move_to_cuda(sample)
    
    return sample

class CudaEnvironment(object):
    def __init__(self):
        cur_device = torch.cuda.current_device()
        prop = torch.cuda.get_device_properties("cuda:{}".format(cur_device))
        self.name = prop.name
        self.major = prop.major
        self.minor = prop.minor
        self.total_memory_in_GB = prop.total_memory / 1024 / 1024 / 1024

    @staticmethod
    def pretty_print_cuda_env_list(cuda_env_list):
        """
        Given a list of CudaEnviorments, pretty print them
        """
        num_workers = len(cuda_env_list)
        center = "CUDA enviroments for all {} workers".format(num_workers)
        banner_len = 40 - len(center) // 2
        first_line = "*" * banner_len + center + "*" * banner_len
        logger.info(first_line)
        for r, env in enumerate(cuda_env_list):
            logger.info(
                "rank {:3d}: ".format(r)
                + "capabilities = {:2d}.{:<2d} ; ".format(env.major, env.minor)
                + "total memory = {:.3f} GB ; ".format(env.total_memory_in_GB)
                + "name = {:40s}".format(env.name)
            )
        logger.info(first_line)

def has_parameters(module):
    try:
        next(module.parameters())
        return True
    except StopIteration:
        return False

@contextmanager
def rename_logger(logger, new_name):
    old_name = logger.name
    if new_name is not None:
        logger.name = new_name
    yield logger
    logger.name = old_name


import math
from dataclasses import dataclass, field
from typing import Any, Dict, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from omegaconf import II

import genhpf.utils.utils as utils
from genhpf.criterions import BaseCriterion, register_criterion
from genhpf.criterions.criterion import CriterionConfig
from genhpf.loggings import meters, metrics
from genhpf.loggings.meters import safe_round


@dataclass
class CrossEntropyConfig(CriterionConfig):
    report_auc: bool = field(
        default=False,
        metadata={
            "help": "whether to report auc. note that this is only available in eval mode and "
            "can cause memory and performance issues if enabled."
        },
    )
    ignore_index: int = II("dataset.ignore_index")


@register_criterion("cross_entropy", dataclass=CrossEntropyConfig)
class CrossEntropy(BaseCriterion):
    def __init__(self, cfg: CrossEntropyConfig):
        super().__init__(cfg)

        if self.task_names is not None and len(self.task_names) > 1:
            raise ValueError(
                "cross_entropy only supports single task training. if you want to train multiple"
                " tasks, use multi_task_criterion instead."
            )

        self.report_auc = cfg.report_auc
        self.ignore_index = cfg.ignore_index

    def compute_loss(
        self, logits: torch.Tensor, targets: torch.Tensor, sample=None, net_output=None, model=None
    ) -> Tuple[torch.Tensor, List[float]]:
        """Compute the loss given the logits and targets from the model."""
        logits = logits.view(-1, logits.size(-1))
        targets = targets.view(-1).long()

        if torch.all(targets == self.ignore_index):
            return logits.new_tensor(0.0), [0.0]

        loss = F.cross_entropy(logits, targets, reduction="sum", ignore_index=self.ignore_index)

        return loss, [loss.detach().item()]

    def get_sample_size(self, sample, targets: torch.Tensor) -> int:
        if "sample_size" in sample:
            sample_size = sample["sample_size"]
        else:
            sample_size = targets.numel()
        return sample_size

    def get_logging_outputs(
        self, logging_output, logits: torch.Tensor, targets: torch.Tensor, sample=None
    ) -> List[Dict[str, Any]]:
        with torch.no_grad():
            logits = logits.view(-1, logits.size(-1))
            targets = targets.view(-1).long()

            valid_indices = torch.where(targets != self.ignore_index)
            if len(valid_indices[0]) == 0:
                return {}

            logits = logits[valid_indices]
            targets = targets[valid_indices]

            preds = logits.argmax(dim=-1)
            count = targets.numel()
            corr = (preds == targets).sum().item()

            logging_output["correct"] = corr
            logging_output["count"] = count

            # report aucs only in eval mode
            if self.report_auc and not self.training:
                probs = torch.sigmoid(logits).view(-1)
                targets = F.one_hot(targets, logits.size(-1)).float().view(-1)

                logging_output["_y_true"] = targets.cpu().numpy()
                logging_output["_y_score"] = probs.cpu().numpy()

        return logging_output

    # def forward(self, model, sample):
    #     net_output = model(**sample['net_input'])
    #     if isinstance(model, DistributedDataParallel):
    #         logits = model.module.get_outputs(
    #                 net_output,
    #                 task=self.args.train_task,
    #                 normalize=False
    #             )
    #         targets = model.module.get_targets(sample, net_output, self.args.train_task)
    #     else:
    #         logits = model.get_outputs(
    #                 net_output,
    #                 task=self.args.train_task,
    #                 normalize=False
    #             )
    #         targets = model.get_targets(sample, net_output, self.args.train_task)

    #     loss_dict = {}
    #     logging_output = {}

    #     if self.args.train_task == 'pretrain' and self.args.pretrain_task in ['mlm', 'spanmlm']:
    #         B, S= targets['input_label'].shape
    #         for victim in self.args.mask_list:
    #             loss = F.cross_entropy(
    #                 logits[victim+'_ids'].view(B*S, -1),
    #                 targets[victim+'_label'].view(-1)
    #             )
    #             loss_dict[victim+'_loss'] = loss

    #             with torch.no_grad():
    #                 preds = torch.argmax(logits[victim+'_ids'], dim=-1).view(-1).detach().cpu()
    #                 target_label = targets[victim+'_label'].view(-1).detach().cpu()
    #                 mask_idcs = (target_label != -100) & (target_label != 0)
    #                 total = mask_idcs.sum()
    #                 correct = (preds[mask_idcs] == target_label[mask_idcs]).sum().float()

    #                 logging_output[victim+'_correct'] = correct
    #                 logging_output[victim+'_total'] = total

    #         loss = sum(loss_dict.values())
    #         sample_size = len(sample)
    #         logging_output['loss'] = loss.item()
    #         logging_output['sample_size'] = sample_size

    #     elif self.args.train_task in ['finetune', 'scratch']:

    #         sample_size = len(targets)
    #         loss = F.cross_entropy(
    #             logits, F.one_hot(
    #                 targets.long(),
    #                 self.multi_label_dict[self.args.pred_src][self.args.pred_target]
    #             ).float().to(logits.device),
    #             reduction=self.ce_reduction_mode
    #         )

    #         logging_output['loss'] = loss.item()
    #         logging_output['sample_size'] = sample_size

    #         with torch.no_grad():
    #             probs = torch.sigmoid(logits).view(-1).detach()
    #             targets = self.mlb.transform(np.expand_dims(targets.view(-1).cpu(), axis=1)).flatten()

    #             logging_output["_y_true"] = targets
    #             logging_output["_y_score"] = probs.cpu().numpy()

    #     return loss, sample_size, logging_output

    @staticmethod
    def reduce_metrics(logging_outputs: List[Dict[str, Any]], prefix: str = None) -> None:
        """Aggregate logging outputs from data parallel training."""
        if prefix is None:
            prefix = ""
        elif prefix is not None and not prefix.endswith("_"):
            prefix = prefix + "_"

        loss_sum = utils.item(sum(log.get("loss", 0) for log in logging_outputs))

        sample_size = utils.item(sum(log.get("sample_size", 0) for log in logging_outputs))

        metrics.log_scalar(f"{prefix}loss", loss_sum / (sample_size or 1) / math.log(2), sample_size, round=3)

        if "_y_true" in logging_outputs[0] and "_y_score" in logging_outputs[0]:
            y_true = np.concatenate([log.get("_y_true", []) for log in logging_outputs])
            y_score = np.concatenate([log.get("_y_score", []) for log in logging_outputs])

            metrics.log_custom(meters.AUCMeter, f"_{prefix}auc", y_score, y_true)

        correct = sum(log.get("correct", 0) for log in logging_outputs)
        metrics.log_scalar(f"_{prefix}correct", correct)

        total = sum(log.get("count", 0) for log in logging_outputs)
        metrics.log_scalar(f"_{prefix}total", total)

        if total > 0:
            metrics.log_derived(
                f"{prefix}accuracy",
                lambda meters: safe_round(meters[f"_{prefix}correct"].sum / meters[f"_{prefix}total"].sum, 5)
                if meters[f"_{prefix}total"].sum > 0
                else float("nan"),
            )

    def post_validate(self, stats, agg, **kwargs):
        for key in agg.keys():
            if key.startswith("_") and key.endswith("auc"):
                stats[key[1:-3] + "auroc"] = agg[key].auroc
                stats[key[1:-3] + "auprc"] = agg[key].auprc

        return stats



import torch
import torch.distributed as dist

import genhpf.utils.distributed_utils as dist_utils

class GatherLayer(torch.autograd.Function):
    """Gather tensors from all process, supporting backward propagation."""

    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)

        group = dist_utils.get_data_parallel_group()
        output = dist_utils.batch_all_gather(input, group=group)

        return tuple(output)
    
    @staticmethod
    def backward(ctx, *grads):
        (input, ) = ctx.saved_tensors
        grad_out = torch.zeros_like(input)
        grad_out[:] = grads[dist.get_rank()]
        return grad_out


import math
from dataclasses import dataclass, field
from typing import List, Tuple

import torch
import torch.nn.functional as F

import genhpf.utils.utils as utils
from genhpf.criterions import BaseCriterion, register_criterion
from genhpf.criterions.criterion import CriterionConfig
from genhpf.loggings import metrics


@dataclass
class SimCLRCriterionConfig(CriterionConfig):
    temp: float = field(default=0.1, metadata={"help": "temperature to divide logits by"})


@register_criterion("simclr_criterion", dataclass=SimCLRCriterionConfig)
class SimCLRCriterion(BaseCriterion):
    def __init__(self, cfg: SimCLRCriterionConfig):
        super().__init__(cfg)

        self.temp = cfg.temp

    def compute_loss(
        self, logits: torch.Tensor, targets: torch.Tensor = None, sample=None, net_output=None, model=None
    ) -> Tuple[torch.Tensor, List[float]]:
        """Compute the loss given the logits and targets from the model."""
        logits = F.normalize(logits, dim=1)  # normalize logits

        bsz = int(logits.shape[0] / 2)

        mask = 1 - torch.eye(bsz * 2, dtype=torch.uint8).to(logits.device)
        pos_ind = (
            torch.arange(bsz * 2).to(logits.device),
            2
            * torch.arange(bsz, dtype=torch.long)
            .unsqueeze(1)
            .repeat(1, 2)
            .view(-1, 1)
            .squeeze()
            .to(logits.device),
        )
        neg_mask = torch.ones((bsz * 2, bsz * 2 - 1), dtype=torch.uint8).to(logits.device)
        neg_mask[pos_ind] = 0

        # Cosine similarity computation
        sim_matrix = torch.matmul(logits, logits.T)  # cosine similarity computation

        # Eliminate similarity between same view
        sim_matrix = torch.masked_select(sim_matrix, mask.bool()).view(sim_matrix.size(0), -1)

        positives = sim_matrix[pos_ind].unsqueeze(1)
        negatives = torch.masked_select(sim_matrix, neg_mask.bool()).view(sim_matrix.size(0), -1)

        logits = torch.cat((positives, negatives), dim=1)
        logits /= self.temp

        target = torch.zeros((logits.size(0),), dtype=torch.long).to(logits.device)

        loss = F.cross_entropy(logits, target, reduction="sum")

        return loss, [loss.detach().item()]

    def get_sample_size(self, sample, targets: torch.Tensor = None) -> int:
        return sample["net_input"]["input_ids"].size(0)

    def get_logging_outputs(self, logging_output, logits, target, sample=None, net_output=None):
        return logging_output

    @staticmethod
    def reduce_metrics(logging_outputs, prefix: str = None) -> None:
        """Aggregate logging outputs from data parallel training."""
        if prefix is None:
            prefix = ""
        elif prefix is not None and not prefix.endswith("_"):
            prefix = prefix + "_"

        loss_sum = utils.item(sum(log.get("loss", 0) for log in logging_outputs))

        sample_size = utils.item(sum(log.get("sample_size", 0) for log in logging_outputs))

        metrics.log_scalar(f"{prefix}loss", loss_sum / (sample_size or 1) / math.log(2), sample_size, round=3)



import glob
import logging
import os

import pandas as pd
import pyspark.sql.functions as F
from ehrs import EHR, register_ehr

logger = logging.getLogger(__name__)


@register_ehr("mimiciv")
class MIMICIV(EHR):
    def __init__(self, cfg):
        super().__init__(cfg)

        self.ehr_name = "mimiciv"

        if self.data_dir is None:
            self.data_dir = os.path.join(self.cache_dir, self.ehr_name)

            if not os.path.exists(self.data_dir):
                logger.info(
                    "Data is not found so try to download from the internet. "
                    "Note that this is a restricted-access resource. "
                    "Please log in to physionet.org with a credentialed user."
                )
                self.download_ehr_from_url(url="https://physionet.org/files/mimiciv/2.0/", dest=self.data_dir)

        logger.info("Data directory is set to {}".format(self.data_dir))

        if self.ccs_path is None:
            self.ccs_path = os.path.join(self.cache_dir, "ccs_multi_dx_tool_2015.csv")

            if not os.path.exists(self.ccs_path):
                logger.info("`ccs_multi_dx_tool_2015.csv` is not found so try to download from the internet.")
                self.download_ccs_from_url(self.cache_dir)

        if self.gem_path is None:
            self.gem_path = os.path.join(self.cache_dir, "icd10cmtoicd9gem.csv")

            if not os.path.exists(self.gem_path):
                logger.info("`icd10cmtoicd9gem.csv` is not found so try to download from the internet.")
                self.download_icdgem_from_url(self.cache_dir)

        if self.ext is None:
            self.ext = self.infer_data_extension()

        self._icustay_fname = "icu/icustays" + self.ext
        self._patient_fname = "hosp/patients" + self.ext
        self._admission_fname = "hosp/admissions" + self.ext
        self._diagnosis_fname = "hosp/diagnoses_icd" + self.ext

        self.tables = [
            {
                "fname": "hosp/labevents" + self.ext,
                "timestamp": "charttime",
                "timeoffsetunit": "abs",
                "exclude": [
                    "labevent_id",
                    "storetime",
                    "subject_id",
                    "specimen_id",
                    "order_provider_id",  # MIMIC-IV-2.2V added this column
                ],
                "code": ["itemid"],
                "desc": ["hosp/d_labitems" + self.ext],
                "desc_key": ["label"],
            },
            {
                "fname": "hosp/prescriptions" + self.ext,
                "timestamp": "starttime",
                "timeoffsetunit": "abs",
                "exclude": [
                    "gsn",
                    "ndc",
                    "subject_id",
                    "pharmacy_id",
                    "poe_id",
                    "poe_seq",
                    "formulary_drug_cd",
                    "stoptime",
                    "order_provider_id",  # MIMIC-IV-2.2V added this column
                ],
            },
            {
                "fname": "icu/inputevents" + self.ext,
                "timestamp": "starttime",
                "timeoffsetunit": "abs",
                "exclude": [
                    "endtime",
                    "storetime",
                    "orderid",
                    "linkorderid",
                    "subject_id",
                    "continueinnextdept",
                    "statusdescription",
                ],
                "code": ["itemid"],
                "desc": ["icu/d_items" + self.ext],
                "desc_key": ["label"],
            },
        ]

        if self.feature == "select":
            extra_exclude_feature_dict = {
                "hosp/labevents"
                + self.ext: ["value", "ref_range_lower", "ref_range_upper", "flag", "priority", "comments"],
                "hosp/prescriptions"
                + self.ext: [
                    "drug_type",
                    "form_rx",
                    "form_val_disp",
                    "form_unit_disp",
                    "doses_per_24_hrs",
                    "route",
                ],
                "icu/inputevents"
                + self.ext: [
                    "amount",
                    "amountuom",
                    "ordercategoryname",
                    "secondaryordercategoryname",
                    "ordercomponenttypedescription",
                    "ordercategorydescription",
                    "patientweight",
                    "totalamount",
                    "totalamountuom",
                    "isopenbag",
                    "originalamount",
                    "originalrate",
                ],
            }

            for table in self.tables:
                if table["fname"] in extra_exclude_feature_dict.keys():
                    exclude_target_list = extra_exclude_feature_dict[table["fname"]]
                    table["exclude"].extend(exclude_target_list)

        if self.emb_type == "codebase":
            feature_types_for_codebase_emb_dict = {
                "hosp/labevents"
                + self.ext: {
                    "numeric_feat": ["value", "valuenum", "ref_range_upper", "ref_range_lower"],
                    "categorical_feat": [],
                    "code_feat": ["itemid"],
                },
                "hosp/prescriptions"
                + self.ext: {
                    "numeric_feat": ["dose_val_rx", "form_val_disp"],
                    "categorical_feat": ["doses_per_24_hrs"],
                    "code_feat": ["drug"],
                },
                "icu/inputevents"
                + self.ext: {
                    "numeric_feat": [
                        "amount",
                        "rate",
                        "patientweight",
                        "totalamount",
                        "originalamount",
                        "originalrate",
                    ],
                    "categorical_feat": ["isopenbag", "continueinnextdept", "cancelreason", "valuecounts"],
                    "code_feat": ["itemid"],
                },
            }

            for table in self.tables:
                if table["fname"] in feature_types_for_codebase_emb_dict.keys():
                    feature_dict = feature_types_for_codebase_emb_dict[table["fname"]]
                    table.update(feature_dict)

        if self.creatinine or self.bilirubin or self.platelets or self.wbc:
            self.task_itemids = {
                "creatinine": {
                    "fname": "hosp/labevents" + self.ext,
                    "timestamp": "charttime",
                    "timeoffsetunit": "abs",
                    "exclude": [
                        "labevent_id",
                        "subject_id",
                        "specimen_id",
                        "storetime",
                        "value",
                        "valueuom",
                        "ref_range_lower",
                        "ref_range_upper",
                        "flag",
                        "priority",
                        "comments",
                    ],
                    "code": ["itemid"],
                    "value": ["valuenum"],
                    "itemid": [50912],
                },
                "bilirubin": {
                    "fname": "hosp/labevents" + self.ext,
                    "timestamp": "charttime",
                    "timeoffsetunit": "abs",
                    "exclude": [
                        "labevent_id",
                        "subject_id",
                        "specimen_id",
                        "storetime",
                        "value",
                        "valueuom",
                        "ref_range_lower",
                        "ref_range_upper",
                        "flag",
                        "priority",
                        "comments",
                    ],
                    "code": ["itemid"],
                    "value": ["valuenum"],
                    "itemid": [50885],
                },
                "platelets": {
                    "fname": "hosp/labevents" + self.ext,
                    "timestamp": "charttime",
                    "timeoffsetunit": "abs",
                    "exclude": [
                        "labevent_id",
                        "subject_id",
                        "specimen_id",
                        "storetime",
                        "value",
                        "valueuom",
                        "ref_range_lower",
                        "ref_range_upper",
                        "flag",
                        "priority",
                        "comments",
                    ],
                    "code": ["itemid"],
                    "value": ["valuenum"],
                    "itemid": [51265],
                },
                "wbc": {
                    "fname": "hosp/labevents" + self.ext,
                    "timestamp": "charttime",
                    "timeoffsetunit": "abs",
                    "exclude": [
                        "labevent_id",
                        "subject_id",
                        "specimen_id",
                        "storetime",
                        "value",
                        "valueuom",
                        "ref_range_lower",
                        "ref_range_upper",
                        "flag",
                        "priority",
                        "comments",
                    ],
                    "code": ["itemid"],
                    "value": ["valuenum"],
                    "itemid": [51300, 51301, 51755],
                },
                "dialysis": {
                    "tables": {
                        "chartevents": {
                            "fname": "icu/chartevents" + self.ext,
                            "timestamp": "charttime",
                            "timeoffsetunit": "abs",
                            "include": ["subject_id", "itemid", "value", "charttime"],
                            "itemid": {
                                "ce": [
                                    226499,
                                    224154,
                                    225183,
                                    227438,
                                    224191,
                                    225806,
                                    225807,
                                    228004,
                                    228005,
                                    228006,
                                    224144,
                                    224145,
                                    224153,
                                    226457,
                                ]
                            },
                        },
                        "inputevents": {
                            "fname": "icu/inputevents" + self.ext,
                            "timestamp": "starttime",
                            "timeoffsetunit": "abs",
                            "include": ["subject_id", "itemid", "amount", "starttime"],
                            "itemid": {"ie": [227536, 227525]},
                        },
                        "procedureevents": {
                            "fname": "icu/procedureevents" + self.ext,
                            "timestamp": "starttime",
                            "timeoffsetunit": "abs",
                            "include": ["subject_id", "itemid", "value", "starttime"],
                            "itemid": {"pe": [225441, 225802, 225803, 225805, 225809, 225955]},
                        },
                    }
                },
            }

        self.disch_map_dict = {
            "ACUTE HOSPITAL": "Other",
            "AGAINST ADVICE": "Other",
            "ASSISTED LIVING": "Other",
            "CHRONIC/LONG TERM ACUTE CARE": "Other",
            "HEALTHCARE FACILITY": "Other",
            "HOME": "Home",
            "HOME HEALTH CARE": "Home",
            "HOSPICE": "Other",
            "IN_ICU_MORTALITY": "IN_ICU_MORTALITY",
            "OTHER FACILITY": "Other",
            "PSYCH FACILITY": "Other",
            "REHAB": "Rehabilitation",
            "SKILLED NURSING FACILITY": "Skilled Nursing Facility",
            "Death": "Death",
        }

        self._icustay_key = "stay_id"
        self._hadm_key = "hadm_id"
        self._patient_key = "subject_id"

        self._determine_first_icu = "INTIME"

    def build_cohorts(self, cached=False):
        icustays = pd.read_csv(os.path.join(self.data_dir, self.icustay_fname))

        icustays = self.make_compatible(icustays)
        self.icustays = icustays

        cohorts = super().build_cohorts(icustays, cached=cached)

        return cohorts

    def prepare_tasks(self, cohorts, spark, cached=False):
        if cached:
            labeled_cohorts = self.load_from_cache(self.ehr_name + ".cohorts.labeled")
            if labeled_cohorts is not None:
                return labeled_cohorts

        labeled_cohorts = super().prepare_tasks(cohorts, spark, cached)

        if self.diagnosis:
            logger.info("Start labeling cohorts for diagnosis prediction.")

            # define diagnosis prediction task
            diagnoses = pd.read_csv(os.path.join(self.data_dir, self.diagnosis_fname))

            diagnoses = self.icd10toicd9(diagnoses)

            ccs_dx = pd.read_csv(self.ccs_path)
            ccs_dx["'ICD-9-CM CODE'"] = ccs_dx["'ICD-9-CM CODE'"].str[1:-1].str.strip()
            ccs_dx["'CCS LVL 1'"] = ccs_dx["'CCS LVL 1'"].str[1:-1]
            lvl1 = {x: int(y) - 1 for _, (x, y) in ccs_dx[["'ICD-9-CM CODE'", "'CCS LVL 1'"]].iterrows()}

            diagnoses["diagnosis"] = diagnoses["icd_code_converted"].map(lvl1)

            diagnoses = diagnoses[(diagnoses["diagnosis"].notnull()) & (diagnoses["diagnosis"] != 14)]
            diagnoses.loc[diagnoses["diagnosis"] >= 14, "diagnosis"] -= 1
            diagnoses = diagnoses.groupby(self.hadm_key)["diagnosis"].agg(lambda x: list(set(x))).to_frame()

            labeled_cohorts = labeled_cohorts.merge(diagnoses, on=self.hadm_key, how="inner")

            logger.info("Done preparing diagnosis prediction for the given cohorts")

            self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled")

        if self.bilirubin or self.platelets or self.creatinine or self.wbc:
            logger.info("Start labeling cohorts for clinical task prediction.")

            labeled_cohorts = spark.createDataFrame(labeled_cohorts)

            if self.bilirubin:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "bilirubin", spark)

            if self.platelets:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "platelets", spark)

            if self.creatinine:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "creatinine", spark)

            if self.wbc:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "wbc", spark)

            # labeled_cohorts = labeled_cohorts.toPandas()

            # self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled.clinical_tasks")

            logger.info("Done preparing clinical task prediction for the given cohorts")

        return labeled_cohorts

    def make_compatible(self, icustays):
        patients = pd.read_csv(os.path.join(self.data_dir, self.patient_fname))
        admissions = pd.read_csv(os.path.join(self.data_dir, self.admission_fname))

        # prepare icustays according to the appropriate format
        icustays = icustays.rename(
            columns={
                "los": "LOS",
                "intime": "INTIME",
                "outtime": "OUTTIME",
            }
        )
        admissions = admissions.rename(
            columns={
                "dischtime": "DISCHTIME",
            }
        )

        icustays = icustays[icustays["first_careunit"] == icustays["last_careunit"]]
        icustays["INTIME"] = pd.to_datetime(icustays["INTIME"], infer_datetime_format=True, utc=True)
        icustays["OUTTIME"] = pd.to_datetime(icustays["OUTTIME"], infer_datetime_format=True, utc=True)

        icustays = icustays.merge(patients, on="subject_id", how="left")
        icustays["AGE"] = icustays["INTIME"].dt.year - icustays["anchor_year"] + icustays["anchor_age"]

        icustays = icustays.merge(
            admissions[[self.hadm_key, "discharge_location", "deathtime", "DISCHTIME"]],
            how="left",
            on=self.hadm_key,
        )

        icustays["discharge_location"].replace("DIED", "Death", inplace=True)
        icustays["DISCHTIME"] = pd.to_datetime(icustays["DISCHTIME"], infer_datetime_format=True, utc=True)

        icustays["IN_ICU_MORTALITY"] = (
            (icustays["INTIME"] < icustays["DISCHTIME"])
            & (icustays["DISCHTIME"] <= icustays["OUTTIME"])
            & (icustays["discharge_location"] == "Death")
        )
        icustays["discharge_location"] = icustays["discharge_location"].map(self.disch_map_dict)
        icustays.rename(columns={"discharge_location": "HOS_DISCHARGE_LOCATION"}, inplace=True)

        icustays["DISCHTIME"] = (icustays["DISCHTIME"] - icustays["INTIME"]).dt.total_seconds() // 60
        icustays["OUTTIME"] = (icustays["OUTTIME"] - icustays["INTIME"]).dt.total_seconds() // 60

        return icustays

    def icd10toicd9(self, dx):
        gem = pd.read_csv(self.gem_path)
        dx_icd_10 = dx[dx["icd_version"] == 10]["icd_code"]

        unique_elem_no_map = set(dx_icd_10) - set(gem["icd10cm"])

        map_cms = dict(zip(gem["icd10cm"], gem["icd9cm"]))
        map_manual = dict.fromkeys(unique_elem_no_map, "NaN")

        for code_10 in map_manual:
            for i in range(len(code_10), 0, -1):
                tgt_10 = code_10[:i]
                if tgt_10 in gem["icd10cm"]:
                    tgt_9 = gem[gem["icd10cm"].str.contains(tgt_10)]["icd9cm"].mode().iloc[0]
                    map_manual[code_10] = tgt_9
                    break

        def icd_convert(icd_version, icd_code):
            if icd_version == 9:
                return icd_code

            elif icd_code in map_cms:
                return map_cms[icd_code]

            elif icd_code in map_manual:
                return map_manual[icd_code]
            else:
                logger.warn("WRONG CODE: " + icd_code)

        dx["icd_code_converted"] = dx.apply(lambda x: icd_convert(x["icd_version"], x["icd_code"]), axis=1)
        return dx

    def clinical_task(self, cohorts, task, spark):
        fname = self.task_itemids[task]["fname"]
        timestamp = self.task_itemids[task]["timestamp"]
        excludes = self.task_itemids[task]["exclude"]
        code = self.task_itemids[task]["code"][0]
        value = self.task_itemids[task]["value"][0]
        itemid = self.task_itemids[task]["itemid"]

        table = spark.read.csv(os.path.join(self.data_dir, fname), header=True)
        table = table.drop(*excludes)
        table = table.filter(F.col(code).isin(itemid)).filter(F.col(value).isNotNull())

        merge = cohorts.join(table, on=self.hadm_key, how="inner")
        merge = merge.withColumn(timestamp, F.to_timestamp(timestamp))

        # Filter Dialysis at here to use abs timestamp & agg by patient_key
        # For Creatinine task, eliminate icus if patient went through dialysis treatment before
        # (obs_size + pred_size / outtime) timestamp
        # see https://github.com/MIT-LCP/mimic-code/blob/main/mimic-iv/concepts/treatment/rrt.sql
        if task == "creatinine":
            dialysis_tables = self.task_itemids["dialysis"]["tables"]

            chartevents = spark.read.csv(
                os.path.join(self.data_dir, "icu/chartevents" + self.ext), header=True
            )
            inputevents = spark.read.csv(
                os.path.join(self.data_dir, "icu/inputevents" + self.ext), header=True
            )
            procedureevents = spark.read.csv(
                os.path.join(self.data_dir, "icu/procedureevents" + self.ext), header=True
            )

            chartevents = chartevents.select(*dialysis_tables["chartevents"]["include"])
            inputevents = inputevents.select(*dialysis_tables["inputevents"]["include"])
            procedureevents = procedureevents.select(*dialysis_tables["procedureevents"]["include"])

            # Filter dialysis related tables with dialysis condition #TODO: check dialysis condition
            ce = chartevents.filter(
                (
                    ((F.col("itemid") == 225965) & (F.col("value") == "In use"))
                    | (F.col("itemid").isin(dialysis_tables["chartevents"]["itemid"]["ce"]))
                    & F.col("value").isNotNull()
                )
            )
            ie = inputevents.filter(
                F.col("itemid").isin(dialysis_tables["inputevents"]["itemid"]["ie"])
            ).filter(F.col("amount") > 0)
            pe = procedureevents.filter(
                F.col("itemid").isin(dialysis_tables["procedureevents"]["itemid"]["pe"])
            ).filter(F.col("value").isNotNull())

            # Extract Dialysis Times!
            def dialysis_time(table, timecolumn):
                return table.withColumn("_DIALYSIS_TIME", F.to_timestamp(timecolumn)).select(
                    self.patient_key, "_DIALYSIS_TIME"
                )

            ce, ie, pe = (
                dialysis_time(ce, "charttime"),
                dialysis_time(ie, "starttime"),
                dialysis_time(pe, "starttime"),
            )
            dialysis = ce.union(ie).union(pe)
            dialysis = dialysis.groupby(self.patient_key).agg(F.min("_DIALYSIS_TIME").alias("_DIALYSIS_TIME"))
            merge = merge.join(dialysis, on=self.patient_key, how="left")
            # Only leave events with no dialysis / before first dialysis
            merge = merge.filter(F.isnull("_DIALYSIS_TIME") | (F.col("_DIALYSIS_TIME") > F.col(timestamp)))
            merge = merge.drop("_DIALYSIS_TIME")

        merge = merge.withColumn(
            timestamp, F.round((F.col(timestamp).cast("long") - F.col("INTIME").cast("long")) / 60)
        )

        # Cohort with events within (obs_size + gap_size) - (obs_size + pred_size)
        merge = merge.filter(((self.obs_size + self.gap_size) * 60) <= F.col(timestamp)).filter(
            ((self.obs_size + self.pred_size) * 60) >= F.col(timestamp)
        )

        # Average value of events
        value_agg = merge.groupBy(self.icustay_key).agg(
            F.mean(value).alias("avg_value")
        )  # TODO: mean/min/max?

        # Labeling
        if task == "bilirubin":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 1.2, 0)
                .when((value_agg.avg_value >= 1.2) & (value_agg.avg_value < 2.0), 1)
                .when((value_agg.avg_value >= 2.0) & (value_agg.avg_value < 6.0), 2)
                .when((value_agg.avg_value >= 6.0) & (value_agg.avg_value < 12.0), 3)
                .when(value_agg.avg_value >= 12.0, 4),
            )
        elif task == "platelets":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value >= 150, 0)
                .when((value_agg.avg_value >= 100) & (value_agg.avg_value < 150), 1)
                .when((value_agg.avg_value >= 50) & (value_agg.avg_value < 100), 2)
                .when((value_agg.avg_value >= 20) & (value_agg.avg_value < 50), 3)
                .when(value_agg.avg_value < 20, 4),
            )

        elif task == "creatinine":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 1.2, 0)
                .when((value_agg.avg_value >= 1.2) & (value_agg.avg_value < 2.0), 1)
                .when((value_agg.avg_value >= 2.0) & (value_agg.avg_value < 3.5), 2)
                .when((value_agg.avg_value >= 3.5) & (value_agg.avg_value < 5), 3)
                .when(value_agg.avg_value >= 5, 4),
            )

        elif task == "wbc":
            # NOTE: unit is mg/L
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 4, 0)
                .when((value_agg.avg_value >= 4) & (value_agg.avg_value <= 12), 1)
                .when((value_agg.avg_value > 12), 2),
            )

        cohorts = cohorts.join(value_agg.select(self.icustay_key, task), on=self.icustay_key, how="left")

        return cohorts

    def infer_data_extension(self) -> str:
        if (
            len(glob.glob(os.path.join(self.data_dir, "hosp", "*.csv.gz"))) == 22
            or len(glob.glob(os.path.join(self.data_dir, "icu", "*.csv.gz"))) == 9
        ):
            ext = ".csv.gz"
        elif (
            len(glob.glob(os.path.join(self.data_dir, "hosp", "*.csv"))) == 22
            or len(glob.glob(os.path.join(self.data_dir, "icu", "*.csv"))) == 9
        ):
            ext = ".csv"
        else:
            raise AssertionError(
                "Provided data directory is not correct. Please check if --data is correct. "
                "--data: {}".format(self.data_dir)
            )

        logger.info("Data extension is set to '{}'".format(ext))

        return ext



import functools
import logging
import glob
import multiprocessing
import os
import re
import shutil
from argparse import ArgumentParser
from bisect import bisect_left, bisect_right
from datetime import datetime
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
import polars as pl
from tqdm import tqdm
from transformers import AutoTokenizer

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

pool_manager = multiprocessing.Manager()
warned_codes = pool_manager.list()


def find_boundary_between(tuples_list, start, end):
    starts = [s for s, e in tuples_list]
    ends = [e for s, e in tuples_list]

    start_index = bisect_left(starts, start)
    end_index = bisect_right(ends, end)
    assert start_index < end_index

    return start_index, end_index


def get_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "root",
        help="path to MEDS dataset. it can be either of directory or the exact file path "
        "with the file extension. if provided with directory, it tries to scan *.csv or "
        "*.parquet files contained in the directory, including sub-directories, to process "
        "all of them.",
    )
    parser.add_argument(
        "--metadata_dir",
        help="path to metadata directory for the input MEDS dataset, which contains codes.parquet",
    )

    parser.add_argument(
        "--birth_code", type=str, default="MEDS_BIRTH", help="string code for the birth event in the dataset."
    )

    parser.add_argument(
        "--cohort",
        type=str,
        help="path to the defined cohort, which must be a result of ACES. it can be either of "
        "directory or the exact file path that has the same extension with the MEDS dataset "
        "to be processed. the file structure of this cohort directory should be the same with "
        "the provided MEDS dataset directory to match each cohort to its corresponding shard "
        "data.",
    )
    parser.add_argument(
        "--cohort_label_name",
        type=str,
        default="boolean_value",
        help="column name in the cohort dataframe to be used for label",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="outputs",
        help="directory to save processed outputs.",
    )
    parser.add_argument(
        "--skip-if-exists",
        action="store_true",
        help="whether or not to skip the processing if the output directory already "
        "exists.",
    )
    parser.add_argument(
        "--rebase",
        action="store_true",
        help="whether or not to rebase the output directory if exists.",
    )
    parser.add_argument(
        "--debug",
        type=bool,
        default=False,
        help="whether or not to enable the debug mode, which forces the script to be run with "
        "only one worker."
    )
    parser.add_argument(
        "--workers",
        metavar="N",
        default=1,
        type=int,
        help="number of parallel workers.",
    )
    parser.add_argument(
        "--max_event_length",
        metavar="N",
        default=128,
        type=int,
        help="maximum number of tokens in an event.",
    )

    parser.add_argument(
        "--mimic_dir",
        default=None,
        help="path to directory for MIMIC-IV database containing hosp/ and icu/ as a subdirectory. "
        "this is used for addressing missing descriptions in the metadata for MIMIC-IV codes.",
    )

    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    root_path = Path(args.root)
    output_dir = Path(args.output_dir)
    metadata_dir = Path(args.metadata_dir)
    mimic_dir = Path(args.mimic_dir) if args.mimic_dir is not None else None

    num_workers = max(args.workers, 1)
    if args.debug:
        num_workers = 1
    else:
        cpu_count = multiprocessing.cpu_count()
        if num_workers > cpu_count:
            logger.warning(
                f"Number of workers (--workers) is greater than the number of available CPUs "
                f"({cpu_count}). Setting the number of workers to {cpu_count}."
            )
            num_workers = cpu_count

    if root_path.is_dir():
        data_paths = glob.glob(str(root_path / "**/*.csv"), recursive=True)
        if len(data_paths) == 0:
            data_paths = glob.glob(str(root_path / "**/*.parquet"), recursive=True)
        if len(data_paths) == 0:
            raise ValueError("Data directory does not contain any supported file formats: .csv or .parquet")
    else:
        data_paths = [root_path]

    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    else:
        if args.rebase:
            shutil.rmtree(output_dir)
            output_dir.mkdir(parents=True)
        elif output_dir.exists():
            if args.skip_if_exists:
                ls = glob.glob(str(output_dir / "**/*"), recursive=True)
                expected_files = []
                for subset in set(os.path.dirname(x) for x in data_paths):
                    expected_files.extend([
                        os.path.join(str(output_dir), os.path.basename(subset), f"{i}.h5")
                        for i in range(num_workers)
                    ])
                if set(expected_files).issubset(set(ls)):
                    logger.info(
                        f"Output directory already contains the expected files. Skipping the "
                        "processing as --skip-if-exists is set. If you want to rebase the directory, "
                        "please run the script with --rebase."
                    )
                    return
            else:
                raise ValueError(
                    f"File exists: '{str(output_dir.resolve())}'. If you want to rebase the "
                    "directory automatically, please run the script with --rebase."
                )

    label_col_name = args.cohort_label_name

    tokenizer = AutoTokenizer.from_pretrained("emilyalsentzer/Bio_ClinicalBERT")

    codes_metadata = pl.read_parquet(metadata_dir / "codes.parquet").to_pandas()
    codes_metadata = codes_metadata.set_index("code")["description"].to_dict()
    # do not allow to use static events or birth event
    birth_code = args.birth_code
    # if birth_code not in codes_metadata:
    #     print(
    #         f'"{birth_code}" is not found in the codes metadata, which may lead to '
    #         "unexpected results since we currently exclude this event from the input data. "
    #     )

    if mimic_dir is not None:
        d_items = pd.read_csv(mimic_dir / "icu" / "d_items.csv.gz")
        d_items["itemid"] = d_items["itemid"].astype("str")
        d_items = d_items.set_index("itemid")["label"].to_dict()
        d_labitems = pd.read_csv(mimic_dir / "hosp" / "d_labitems.csv.gz")
        d_labitems["itemid"] = d_labitems["itemid"].astype("str")
        d_labitems = d_labitems.set_index("itemid")["label"].to_dict()
    else:
        d_items = None
        d_labitems = None

    max_event_length = args.max_event_length

    progress_bar = tqdm(data_paths, total=len(data_paths))
    for data_path in progress_bar:
        progress_bar.set_description(str(data_path))

        data_path = Path(data_path)
        subdir = data_path.relative_to(root_path).parent
        if data_path.suffix == ".csv":
            data = pl.scan_csv(data_path)
        elif data_path.suffix == ".parquet":
            data = pl.scan_parquet(data_path)
        else:
            raise ValueError(f"Unsupported file format: {data_path.suffix}")

        data = data.with_columns(
            pl.when(pl.col("code") == birth_code).then(None).otherwise(pl.col("time")).alias("time")
        )
        data = data.drop_nulls(subset=["subject_id", "time"])

        cohort_path = Path(args.cohort) / subdir / data_path.name

        if cohort_path.suffix == ".csv":
            cohort = pl.scan_csv(cohort_path)
        elif cohort_path.suffix == ".parquet":
            cohort = pl.scan_parquet(cohort_path)
        else:
            raise ValueError(f"Unsupported file format: {cohort_path.suffix}")

        cohort = cohort.drop_nulls(label_col_name)
        cohort = cohort.unique()

        cohort = cohort.select(
            [
                pl.col("subject_id"),
                pl.col(label_col_name),
                # pl.col("input.end_summary").struct.field("timestamp_at_start").alias("starttime"),
                pl.col("prediction_time").alias("endtime"),
            ]
        )
        cohort = (
            cohort.group_by("subject_id", maintain_order=True)
            .agg(pl.col(["endtime", label_col_name]))
            .collect()
        )  # omitted "starttime"
        cohort_dict = {
            x["subject_id"]: {
                # "starttime": x["starttime"],
                "endtime": x["endtime"],
                "label": x[label_col_name],
            }
            for x in cohort.iter_rows(named=True)
        }

        def extract_cohort(row):
            subject_id = row["subject_id"]
            time = row["time"]
            if subject_id not in cohort_dict:
                # return {"cohort_start": None, "cohort_end": None, "cohort_label": None}
                return {"cohort_end": None, "cohort_label": None}

            cohort_criteria = cohort_dict[subject_id]
            # starts = cohort_criteria["starttime"]
            ends = cohort_criteria["endtime"]
            labels = cohort_criteria["label"]

            # for start, end, label in zip(starts, ends, labels):
            #     if start <= time and time <= end:
            #         return {"cohort_start": start, "cohort_end": end, "cohort_label": label}

            # assume it is possible that each event goes into multiple different cohorts
            cohort_ends = []
            cohort_labels = []
            for end, label in zip(ends, labels):
                if time <= end:
                    # return {"cohort_start": start, "cohort_end": end, "cohort_label": label}
                    cohort_ends.append(end)
                    cohort_labels.append(label)

            if len(cohort_ends) > 0:
                return {"cohort_end": cohort_ends, "cohort_label": cohort_labels}
            else:
                # return {"cohort_start": None, "cohort_end": None, "cohort_label": None}
                return {"cohort_end": None, "cohort_label": None}

        data = data.group_by(["subject_id", "time"], maintain_order=True).agg(pl.all())
        data = (
            data.with_columns(
                pl.struct(["subject_id", "time"])
                .map_elements(
                    extract_cohort,
                    return_dtype=pl.Struct(
                        {
                            "cohort_end": pl.List(pl.Datetime()),
                            "cohort_label": pl.List(pl.Boolean),
                        }
                    ),
                )
                .alias("cohort_criteria")
            )
            .unnest("cohort_criteria")
            .collect()
        )

        data = data.drop_nulls("cohort_label")

        data = data.with_columns(pl.col("time").dt.strftime("%Y-%m-%d %H:%M:%S").cast(pl.List(str)))
        data = data.with_columns(
            pl.col("time").list.sample(n=pl.col("code").list.len(), with_replacement=True)
        )

        if str(subdir) != ".":
            output_name = str(subdir)
        else:
            output_name = data_path.stem

        if not os.path.exists(output_dir / output_name):
            os.makedirs(output_dir / output_name)

        with open(str(output_dir / (output_name + ".tsv")), "a") as manifest_f:
            if os.path.getsize(output_dir / (output_name + ".tsv")) == 0:
                manifest_f.write(f"{output_dir}/{output_name}\n")

            must_have_columns = [
                "subject_id",
                "cohort_end",
                "cohort_label",
                "time",
                "code",
                "numeric_value",
            ]
            rest_of_columns = [x for x in data.columns if x not in must_have_columns]
            column_name_idcs = {col: i for i, col in enumerate(data.columns)}

            meds_to_remed_partial = functools.partial(
                meds_to_remed,
                tokenizer,
                rest_of_columns,
                column_name_idcs,
                codes_metadata,
                output_dir,
                output_name,
                num_workers,
                d_items,
                d_labitems,
                warned_codes,
                max_event_length,
            )

            # meds --> remed
            logger.info(f"Start processing {data_path}")
            if num_workers <= 1:
                length_per_subject_gathered = [meds_to_remed_partial(data)]
                del data
            else:
                subject_ids = data["subject_id"].unique().to_list()
                n = num_workers
                subject_id_chunks = [subject_ids[i::n] for i in range(n)]
                data_chunks = []
                for subject_id_chunk in subject_id_chunks:
                    data_chunks.append(data.filter(pl.col("subject_id").is_in(subject_id_chunk)))
                del data

                num_valid_data_chunks = sum(map(lambda x: len(x) > 0, data_chunks))
                if num_valid_data_chunks < num_workers:
                    raise ValueError(
                        "Number of valid data chunks (= number of unique subjects) were smaller "
                        "than the specified num workers (--workers) due to the small size of data. "
                        "Consider reducing the number of workers."
                    )

                pool = multiprocessing.get_context("spawn").Pool(processes=num_workers)
                # the order is preserved
                length_per_subject_gathered = pool.map(meds_to_remed_partial, data_chunks)
                pool.close()
                pool.join()
                del data_chunks

            if len(length_per_subject_gathered) != num_workers:
                raise ValueError(
                    "Number of processed workers were smaller than the specified num workers "
                    "(--workers) due to the small size of data. Consider reducing the number of "
                    "workers."
                )

            for length_per_subject in length_per_subject_gathered:
                for subject_id, (length, shard_id) in length_per_subject.items():
                    manifest_f.write(f"{subject_id}\t{length}\t{shard_id}\n")


def meds_to_remed(
    tokenizer,
    rest_of_columns,
    column_name_idcs,
    codes_metadata,
    output_dir,
    output_name,
    num_shards,
    d_items,
    d_labitems,
    warned_codes,
    max_event_length,
    df_chunk,
):
    code_matching_pattern = re.compile(r"\d+")

    def meds_to_remed_unit(row):
        events = []
        digit_offsets = []
        col_name_offsets = []
        for event_index in range(len(row[column_name_idcs["code"]])):
            event = ""
            digit_offset = []
            col_name_offset = []
            for col_name in ["code", "numeric_value"] + rest_of_columns:
                # do not process something like "icustay_id" or "hadm_id"
                if "id" in col_name:
                    continue

                col_event = row[column_name_idcs[col_name]][event_index]
                if col_event is not None:
                    col_event = str(col_event)
                    if col_name == "code":
                        if col_event in codes_metadata and codes_metadata[col_event] != "":
                            col_event = codes_metadata[col_event]
                        else:
                            do_break = False
                            items = col_event.split("//")
                            is_code = [bool(code_matching_pattern.fullmatch(item)) for item in items]
                            if True in is_code:
                                if d_items is not None and d_labitems is not None:
                                    code_idx = is_code.index(True)
                                    code = items[code_idx]

                                    if code in d_items:
                                        desc = d_items[code]
                                    elif code in d_labitems:
                                        desc = d_labitems[code]
                                    else:
                                        do_break = True

                                    if not do_break:
                                        items[code_idx] = desc
                                        col_event = "//".join(items)
                                else:
                                    do_break = True

                            if do_break and col_event not in warned_codes:
                                warned_codes.append(col_event)
                                logger.warning(
                                    "The dataset contains some codes that are not specified in "
                                    "the codes metadata, which may not be intended. Note that we "
                                    f"process this code as it is for now: {col_event}."
                                )
                    else:
                        col_event = re.sub(
                            r"\d*\.\d+",
                            lambda x: str(round(float(x.group(0)), 4)),
                            col_event,
                        )
                        event_offset = len(event) + len(col_name) + 1
                        digit_offset_tmp = [
                            g.span() for g in re.finditer(r"([0-9]+([.][0-9]*)?|[0-9]+|\.+)", col_event)
                        ]

                        internal_offset = 0
                        for start, end in digit_offset_tmp:
                            digit_offset.append(
                                (
                                    event_offset + start + internal_offset,
                                    event_offset + end + (end - start) * 2 + internal_offset,
                                )
                            )
                            internal_offset += (end - start) * 2

                        col_event = re.sub(r"([0-9\.])", r" \1 ", col_event)

                    col_name_offset.append((len(event), len(event) + len(col_name)))
                    event += " " + col_name + " " + col_event
            if len(event) > 0:
                events.append(event[1:])
                digit_offsets.append(digit_offset)
                col_name_offsets.append(col_name_offset)

        tokenized_events = tokenizer(
            events,
            add_special_tokens=True,
            padding="max_length",
            max_length=max_event_length,
            truncation=True,
            return_tensors="np",
            return_token_type_ids=False,
            return_attention_mask=True,
            return_offsets_mapping=True,
        )
        lengths_before_padding = tokenized_events["attention_mask"].sum(axis=1)

        input_ids = tokenized_events["input_ids"]
        dpe_ids = np.zeros(input_ids.shape, dtype=int)
        for i, digit_offset in enumerate(digit_offsets):
            for start, end in digit_offset:
                start_index, end_index = find_boundary_between(
                    tokenized_events[i].offsets[: lengths_before_padding[i] - 1],
                    start,
                    end,
                )

                # define dpe ids for digits found
                num_digits = end_index - start_index
                # 119: token id for "."
                num_decimal_points = (input_ids[i][start_index:end_index] == 119).sum()

                # integer without decimal point
                # e.g., for "1 2 3 4 5", assign [10, 9, 8, 7, 6]
                if num_decimal_points == 0:
                    dpe_ids[i][start_index:end_index] = list(range(num_digits + 5, 5, -1))
                # floats
                # e.g., for "1 2 3 4 5 . 6 7 8 9", assign [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]
                elif num_decimal_points == 1:
                    num_decimals = (
                        num_digits
                        - np.where(input_ids[i][start_index:end_index] == 119)[0][0]  # 119: token id for "."
                    )
                    dpe_ids[i][start_index:end_index] = list(
                        range(num_digits + 5 - num_decimals, 5 - num_decimals, -1)
                    )
                # 1 > decimal points where we cannot define dpe ids
                else:
                    continue
        # define type ids
        # for column names: 2
        # for column values (contents): 3
        # for CLS tokens: 5
        # for SEP tokens: 6
        type_ids = np.zeros(input_ids.shape, dtype=int)
        type_ids[:, 0] = 5  # CLS tokens
        for i, col_name_offset in enumerate(col_name_offsets):
            type_ids[i][lengths_before_padding[i] - 1] = 6  # SEP tokens
            # fill with type ids for column values
            type_ids[i][1 : lengths_before_padding[i] - 1] = 3
            for start, end in col_name_offset:
                start_index, end_index = find_boundary_between(
                    tokenized_events[i].offsets[1 : lengths_before_padding[i] - 1],
                    start,
                    end,
                )
                # the first offset is always (0, 0) for CLS token, so we adjust it
                start_index += 1
                end_index += 1
                # finally replace with type ids for column names
                type_ids[i][start_index:end_index] = 2

        return np.stack([input_ids, type_ids, dpe_ids], axis=1).astype(np.uint16)

    events_data = []
    worker_id = multiprocessing.current_process().name.split("-")[-1]
    if worker_id == "MainProcess":
        worker_id = 0
    else:
        # worker_id is incremental for every generated pool, so divide with num_shards
        worker_id = (int(worker_id) - 1) % num_shards  # 1-based -> 0-based indexing
    if worker_id == 0:
        progress_bar = tqdm(df_chunk.iter_rows(), total=len(df_chunk))
        progress_bar.set_description(f"Processing from worker-{worker_id}:")
    else:
        progress_bar = df_chunk.iter_rows()

    for row in progress_bar:
        events_data.append(meds_to_remed_unit(row))
    data_length = list(map(len, events_data))
    data_index_offset = np.zeros(len(data_length), dtype=np.int64)
    data_index_offset[1:] = np.cumsum(data_length[:-1])
    data_index = pl.Series(
        "data_index",
        map(
            lambda x: [data_index_offset[x] + y for y in range(data_length[x])],
            range(len(data_length)),
        ),
    )
    events_data = np.concatenate(events_data)

    df_chunk = df_chunk.select(["subject_id", "cohort_end", "cohort_label", "time"])
    df_chunk = df_chunk.insert_column(4, data_index)
    df_chunk = df_chunk.explode(["cohort_end", "cohort_label"])
    df_chunk = df_chunk.group_by(
        # ["subject_id", "cohort_start", "cohort_end", "cohort_label"], maintain_order=True
        ["subject_id", "cohort_end", "cohort_label"],
        maintain_order=True,
    ).agg(pl.all())

    df_chunk = df_chunk.sort(by=["subject_id", "cohort_end"])
    # regard {subject_id} as {cohort_id}: {subject_id}_{cohort_number}
    df_chunk = df_chunk.with_columns(pl.col("subject_id").cum_count().over("subject_id").alias("suffix"))
    df_chunk = df_chunk.with_columns(
        (pl.col("subject_id").cast(str) + "_" + pl.col("suffix").cast(str)).alias("subject_id")
    )
    # data = data.drop("suffix", "cohort_start", "cohort_end")
    df_chunk = df_chunk.drop("suffix", "cohort_end")

    length_per_subject = {}
    progress_bar = tqdm(
        df_chunk.iter_rows(),
        total=len(df_chunk),
        desc=f"Writing data from worker-{worker_id}:",
    )

    for sample in progress_bar:
        with h5py.File(str(output_dir / output_name / f"{worker_id}.h5"), "a") as f:
            if "ehr" in f:
                result = f["ehr"]
            else:
                result = f.create_group("ehr")

            sample_result = result.create_group(sample[0])

            data_indices = np.concatenate(sample[3])
            data = events_data[data_indices]
            sample_result.create_dataset("hi", data=data, dtype="i2", compression="lzf", shuffle=True)

            times = np.concatenate(sample[2])
            times = [datetime.strptime(x, "%Y-%m-%d %H:%M:%S") for x in times]
            times = np.cumsum(np.diff(times))
            times = list(map(lambda x: round(x.total_seconds() / 60), times))
            times = np.array([0] + times)

            sample_result.create_dataset("time", data=times, dtype="i")
            sample_result.create_dataset("label", data=int(sample[1]))

            length_per_subject[sample[0]] = (len(times), worker_id)
    del df_chunk

    return length_per_subject


if __name__ == "__main__":
    main()



from .gather_layer import GatherLayer
from .grad_multiply import GradMultiply
from .gumbel_vector_quantizer import GumbelVectorQuantizer
from .identity_layer import Identity
from .layer_norm import LayerNorm
from .positional_encoding import PositionalEncoding

__all__ = [
    "Identity",
    "GatherLayer",
    "GradMultiply",
    "GumbelVectorQuantizer",
    "LayerNorm",
    "PositionalEncoding",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from dataclasses import _MISSING_TYPE
from hydra.core.config_store import ConfigStore
from omegaconf import DictConfig, OmegaConf

from genhpf.configs import Config

logger = logging.getLogger(__name__)

def hydra_init(cfg_name="config") -> None:
    cs = ConfigStore.instance()
    cs.store(name=cfg_name, node=Config)

    for k in Config.__dataclass_fields__:
        v = Config.__dataclass_fields__[k].default
        if isinstance(v, _MISSING_TYPE):
            v = Config.__dataclass_fields__[k].default_factory
            if not isinstance(v, _MISSING_TYPE):
                v = v()
        try:
            cs.store(name = k, node = v)
        except BaseException:
            logger.error(f"{k} - {v}")
            raise

def add_defaults(cfg: DictConfig) -> None:
    """This function adds default values that are stored in dataclasses that hydra doesn't know about """

    from genhpf.criterions import CRITERION_DATACLASS_REGISTRY
    from genhpf.models import MODEL_DATACLASS_REGISTRY
    from genhpf.configs.utils import merge_with_parent
    from typing import Any

    OmegaConf.set_struct(cfg, False)

    for k, v in Config.__dataclass_fields__.items():
        field_cfg = cfg.get(k)
        if field_cfg is not None and v.type == Any:
            dc = None

            if isinstance(field_cfg, str):
                field_cfg = DictConfig({"_name": field_cfg})
                field_cfg.__dict__["_parent"] = field_cfg.__dict__["_parent"]

            name = field_cfg.get("_name")

            if k == "model":
                dc = MODEL_DATACLASS_REGISTRY.get(name)
            elif k == "criterion":
                dc = CRITERION_DATACLASS_REGISTRY.get(name)
            
            if dc is not None:
                cfg[k] = merge_with_parent(dc, field_cfg)


import glob
import logging
import os
from collections import Counter

import pandas as pd
import pyspark.sql.functions as F
import treelib
from ehrs import EHR, register_ehr

logger = logging.getLogger(__name__)


@register_ehr("eicu")
class eICU(EHR):
    def __init__(self, cfg):
        super().__init__(cfg)

        self.ehr_name = "eicu"

        if self.data_dir is None:
            self.data_dir = os.path.join(self.cache_dir, self.ehr_name)

            if not os.path.exists(self.data_dir):
                logger.info(
                    "Data is not found so try to download from the internet. "
                    "Note that this is a restricted-access resource. "
                    "Please log in to physionet.org with a credentialed user."
                )
                self.download_ehr_from_url(
                    url="https://physionet.org/files/eicu-crd/2.0/", dest=self.data_dir
                )

        logger.info("Data directory is set to {}".format(self.data_dir))

        if self.ccs_path is None:
            self.ccs_path = os.path.join(self.cache_dir, "ccs_multi_dx_tool_2015.csv")

            if not os.path.exists(self.ccs_path):
                logger.info("`ccs_multi_dx_tool_2015.csv` is not found so try to download from the internet.")
                self.download_ccs_from_url(self.cache_dir)

        if self.gem_path is None:
            self.gem_path = os.path.join(self.cache_dir, "icd10cmtoicd9gem.csv")

            if not os.path.exists(self.gem_path):
                logger.info("`icd10cmtoicd9gem.csv` is not found so try to download from the internet.")
                self.download_icdgem_from_url(self.cache_dir)

        if self.ext is None:
            self.ext = self.infer_data_extension()

        self._icustay_fname = "patient" + self.ext
        self._diagnosis_fname = "diagnosis" + self.ext

        self.tables = [
            {
                "fname": "lab" + self.ext,
                "timestamp": "labresultoffset",
                "timeoffsetunit": "min",
                "exclude": ["labid", "labresultrevisedoffset"],
            },
            {
                "fname": "medication" + self.ext,
                "timestamp": "drugstartoffset",
                "timeoffsetunit": "min",
                "exclude": [
                    "drugorderoffset",
                    "drugstopoffset",
                    "medicationid",
                    "gtc",
                    "drughiclseqno",
                    "drugordercancelled",
                ],
            },
            {
                "fname": "infusionDrug" + self.ext,
                "timestamp": "infusionoffset",
                "timeoffsetunit": "min",
                "exclude": ["infusiondrugid"],
            },
        ]

        if self.feature == "select":
            extra_exclude_feature_dict = {
                "lab"
                + self.ext: [
                    "labtypeid",
                    "labresulttext",
                    "labmeasurenameinterface",
                    "labresultrevisedoffset",
                ],
                "medication" + self.ext: ["drugivadmixture" "frequency", "loadingdose", "prn"],
                "infusionDrug" + self.ext: ["patientweight", "volumeoffluid", "drugrate", "drugamount"],
            }

            for table in self.tables:
                if table["fname"] in extra_exclude_feature_dict.keys():
                    exclude_target_list = extra_exclude_feature_dict[table["fname"]]
                    table["exclude"].extend(exclude_target_list)

        if self.emb_type == "codebase":
            feature_types_for_codebase_emb_dict = {
                "lab"
                + self.ext: {
                    "numeric_feat": ["labresult", "labresulttext"],
                    "categorical_feat": ["labtypeid"],
                    "code_feat": ["labname"],
                },
                "medication"
                + self.ext: {
                    "numeric_feat": ["dosage"],
                    "categorical_feat": ["drugordercancelled", "drugivadmixture"],
                    "code_feat": ["drugname"],
                },
                "infusionDrug"
                + self.ext: {
                    "numeric_feat": [
                        "drugrate",
                        "infusionrate",
                        "drugamount",
                        "volumeoffluid",
                        "patientweight",
                    ],
                    "categorical_feat": [],
                    "code_feat": ["drugname"],
                },
            }

            for table in self.tables:
                if table["fname"] in feature_types_for_codebase_emb_dict.keys():
                    feature_dict = feature_types_for_codebase_emb_dict[table["fname"]]
                    table.update(feature_dict)

        if self.creatinine or self.bilirubin or self.platelets or self.wbc:
            self.task_itemids = {
                "creatinine": {
                    "fname": "lab" + self.ext,
                    "timestamp": "labresultoffset",
                    "timeoffsetunit": "min",
                    "exclude": [
                        "labtypeid",
                        "labresulttext",
                        "labmeasurenamesystem",
                        "labmeasurenameinterface",
                        "labresultrevisedoffset",
                    ],
                    "code": ["labname"],
                    "value": ["labresult"],
                    "itemid": ["creatinine"],
                },
                "bilirubin": {
                    "fname": "lab" + self.ext,
                    "timestamp": "labresultoffset",
                    "timeoffsetunit": "min",
                    "exclude": [
                        "labtypeid",
                        "labresulttext",
                        "labmeasurenamesystem",
                        "labmeasurenameinterface",
                        "labresultrevisedoffset",
                    ],
                    "code": ["labname"],
                    "value": ["labresult"],
                    "itemid": ["total bilirubin"],
                },
                "platelets": {
                    "fname": "lab" + self.ext,
                    "timestamp": "labresultoffset",
                    "timeoffsetunit": "min",
                    "exclude": [
                        "labtypeid",
                        "labresulttext",
                        "labmeasurenamesystem",
                        "labmeasurenameinterface",
                        "labresultrevisedoffset",
                    ],
                    "code": ["labname"],
                    "value": ["labresult"],
                    "itemid": ["platelets x 1000"],
                },
                "wbc": {
                    "fname": "lab" + self.ext,
                    "timestamp": "labresultoffset",
                    "timeoffsetunit": "min",
                    "exclude": [
                        "labtypeid",
                        "labresulttext",
                        "labmeasurenamesystem",
                        "labmeasurenameinterface",
                        "labresultrevisedoffset",
                    ],
                    "code": ["labname"],
                    "value": ["labresult"],
                    "itemid": ["WBC x 1000"],
                },
                "dialysis": {
                    "fname": "intakeOutput" + self.ext,
                    "timestamp": "intakeoutputoffset",
                    "timeoffsetunit": "min",
                    "exclude": [
                        "intakeoutputid",
                        "intaketotal",
                        "outputtotal",
                        "nettotal",
                        "intakeoutputentryoffset",
                    ],
                    "code": ["dialysistotal"],
                    "value": [],
                    "itemid": [],
                },
            }

        self.disch_map_dict = {
            "Home": "Home",
            "IN_ICU_MORTALITY": "IN_ICU_MORTALITY",
            "Nursing Home": "Other",
            "Other": "Other",
            "Other External": "Other",
            "Other Hospital": "Other",
            "Rehabilitation": "Rehabilitation",
            "Skilled Nursing Facility": "Skilled Nursing Facility",
            "Death": "Death",
        }

        self._icustay_key = "patientunitstayid"
        self._hadm_key = "patienthealthsystemstayid"
        self._patient_key = "uniquepid"

        self._determine_first_icu = "unitvisitnumber"

    def build_cohorts(self, cached=False):
        icustays = pd.read_csv(os.path.join(self.data_dir, self.icustay_fname))

        icustays = self.make_compatible(icustays)
        self.icustays = icustays

        cohorts = super().build_cohorts(icustays, cached=cached)

        return cohorts

    def prepare_tasks(self, cohorts, spark, cached=False):
        if cohorts is None and cached:
            labeled_cohorts = self.load_from_cache(self.ehr_name + ".cohorts.labeled")
            if labeled_cohorts is not None:
                return labeled_cohorts

        labeled_cohorts = super().prepare_tasks(cohorts, spark, cached)

        if self.diagnosis:
            logger.info("Start labeling cohorts for diagnosis prediction.")

            str2cat = self.make_dx_mapping()
            dx = pd.read_csv(os.path.join(self.data_dir, self.diagnosis_fname))
            dx = dx.merge(cohorts[[self.icustay_key, self.hadm_key]], on=self.icustay_key)
            dx["diagnosis"] = dx["diagnosisstring"].map(lambda x: str2cat.get(x, -1))
            # Ignore Rare Class(14)
            dx = dx[(dx["diagnosis"] != -1) & (dx["diagnosis"] != 14)]
            dx.loc[dx["diagnosis"] >= 14, "diagnosis"] -= 1
            dx = dx.groupby(self.hadm_key)["diagnosis"].agg(lambda x: list(set(x))).to_frame()

            labeled_cohorts = labeled_cohorts.merge(dx, on=self.hadm_key, how="left")
            labeled_cohorts["diagnosis"] = labeled_cohorts["diagnosis"].apply(
                lambda x: [] if type(x) is not list else x
            )
            # NaN case in diagnosis -> []

        logger.info("Done preparing diagnosis prediction for the given cohorts")

        self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled")

        if self.bilirubin or self.platelets or self.creatinine or self.wbc:
            logger.info("Start labeling cohorts for clinical task prediction.")

            labeled_cohorts = spark.createDataFrame(labeled_cohorts)

            if self.bilirubin:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "bilirubin", spark)

            if self.platelets:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "platelets", spark)

            if self.creatinine:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "creatinine", spark)

            if self.wbc:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "wbc", spark)

            # self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled.clinical_tasks")

            logger.info("Done preparing clinical task prediction for the given cohorts")

        return labeled_cohorts

    def make_compatible(self, icustays):
        icustays.loc[:, "LOS"] = icustays["unitdischargeoffset"] / 60 / 24
        icustays.dropna(subset=["age"], inplace=True)
        icustays["AGE"] = icustays["age"].replace("> 89", 300).astype(int)

        # hacks for compatibility with other ehrs
        icustays["INTIME"] = 0
        icustays.rename(columns={"unitdischargeoffset": "OUTTIME"}, inplace=True)
        # DEATHTIME
        # icustays["DEATHTIME"] = np.nan
        # is_discharged_in_icu = icustays["unitdischargestatus"] == "Expired"
        # icustays.loc[is_discharged_in_icu, "DEATHTIME"] = (
        #     icustays.loc[is_discharged_in_icu, "OUTTIME"]
        # )
        # is_discharged_in_hos = (
        #     (icustays["unitdischargestatus"] != "Expired")
        #     & (icustays["hospitaldischargestatus"] == "Expired")
        # )
        # icustays.loc[is_discharged_in_hos, "DEATHTIME"] = (
        #     icustays.loc[is_discharged_in_hos, "OUTTIME"]
        # ) + 1

        icustays.rename(columns={"hospitaldischargeoffset": "DISCHTIME"}, inplace=True)

        icustays["IN_ICU_MORTALITY"] = icustays["unitdischargestatus"] == "Expired"
        icustays["hospitaldischargelocation"] = icustays["hospitaldischargelocation"].map(self.disch_map_dict)
        icustays.rename(columns={"hospitaldischargelocation": "HOS_DISCHARGE_LOCATION"}, inplace=True)

        return icustays

    def make_dx_mapping(self):
        diagnosis = pd.read_csv(os.path.join(self.data_dir, self.diagnosis_fname))
        ccs_dx = pd.read_csv(self.ccs_path)
        gem = pd.read_csv(self.gem_path)

        diagnosis = diagnosis[["diagnosisstring", "icd9code"]]

        # 1 to 1 matching btw str and code
        # STR: diagnosisstring, CODE: icd9/10 code, CAT:category
        # 1 str -> multiple code -> one cat

        # 1. make str -> code dictonary
        str2code = diagnosis.dropna(subset=["icd9code"])
        str2code = str2code.groupby("diagnosisstring").first().reset_index()
        str2code["icd9code"] = str2code["icd9code"].str.split(",")
        str2code = str2code.explode("icd9code")
        str2code["icd9code"] = str2code["icd9code"].str.replace(".", "", regex=False)
        # str2code = dict(zip(notnull_dx["diagnosisstring"], notnull_dx["icd9code"]))
        # 이거 하면 dxstring duplicated 자동 제거됨 ->x

        ccs_dx["'ICD-9-CM CODE'"] = ccs_dx["'ICD-9-CM CODE'"].str[1:-1].str.strip()
        ccs_dx["'CCS LVL 1'"] = ccs_dx["'CCS LVL 1'"].str[1:-1].astype(int) - 1
        icd2cat = dict(zip(ccs_dx["'ICD-9-CM CODE'"], ccs_dx["'CCS LVL 1'"]))

        # 2. if code is not icd9, convert it to icd9
        str2code_icd10 = str2code[str2code["icd9code"].isin(icd2cat.keys())]

        map_cms = dict(zip(gem["icd10cm"], gem["icd9cm"]))
        map_manual = dict.fromkeys(set(str2code_icd10["icd9code"]) - set(gem["icd10cm"]), "NaN")

        for code_10 in map_manual:
            for i in range(len(code_10), 0, -1):
                tgt_10 = code_10[:i]
                if tgt_10 in gem["icd10cm"]:
                    tgt_9 = gem[gem["icd10cm"].str.contains(tgt_10)]["icd9cm"].mode().iloc[0]
                    map_manual[code_10] = tgt_9
                    break
        icd102icd9 = {**map_cms, **map_manual}

        # 3. Convert Available Strings to category
        str2cat = {}
        for _, row in str2code.iterrows():
            k, v = row
            if v in icd2cat.keys():
                cat = icd2cat[v]
                if k in str2cat.keys() and str2cat[k] != cat:
                    logger.warning(f"{k} has multiple categories{cat, str2cat[k]}")
                str2cat[k] = icd2cat[v]
            elif v in icd102icd9.keys():
                cat = icd2cat[icd102icd9[v]]
                if k in str2cat.keys() and str2cat[k] != cat:
                    logger.warning(f"{k} has multiple categories{cat, str2cat[k]}")
                str2cat[k] = icd2cat[icd102icd9[v]]

        # 4. If no available category by mapping(~25%), use diagnosisstring hierarchy

        # Make tree structure
        tree = treelib.Tree()
        tree.create_node("root", "root")
        for dx, cat in str2cat.items():
            dx = dx.split("|")
            if not tree.contains(dx[0]):
                tree.create_node(-1, dx[0], parent="root")
            for i in range(2, len(dx)):
                if not tree.contains("|".join(dx[:i])):
                    tree.create_node(-1, "|".join(dx[:i]), parent="|".join(dx[: i - 1]))
            if not tree.contains("|".join(dx)):
                tree.create_node(cat, "|".join(dx), parent="|".join(dx[:-1]))

        # Update non-leaf nodes with majority vote
        nid_list = list(tree.expand_tree(mode=treelib.Tree.DEPTH))
        nid_list.reverse()
        for nid in nid_list:
            if tree.get_node(nid).is_leaf():
                continue
            elif tree.get_node(nid).tag == -1:
                tree.get_node(nid).tag = Counter([child.tag for child in tree.children(nid)]).most_common(1)[
                    0
                ][0]

        # Evaluate dxs without category
        unmatched_dxs = set(diagnosis["diagnosisstring"]) - set(str2cat.keys())
        for dx in unmatched_dxs:
            dx = dx.split("|")
            # Do not go to root level(can add noise)
            for i in range(len(dx) - 1, 1, -1):
                if tree.contains("|".join(dx[:i])):
                    str2cat["|".join(dx)] = tree.get_node("|".join(dx[:i])).tag
                    break

        return str2cat

    def clinical_task(self, cohorts, task, spark):
        fname = self.task_itemids[task]["fname"]
        timestamp = self.task_itemids[task]["timestamp"]
        excludes = self.task_itemids[task]["exclude"]
        code = self.task_itemids[task]["code"][0]
        value = self.task_itemids[task]["value"][0]
        itemid = self.task_itemids[task]["itemid"]

        table = spark.read.csv(os.path.join(self.data_dir, fname), header=True)
        table = table.drop(*excludes)
        table = table.filter(F.col(code).isin(itemid)).filter(F.col(value).isNotNull())

        merge = cohorts.join(table, on=self.icustay_key, how="inner")

        if task == "creatinine":
            patient = spark.read.csv(os.path.join(self.data_dir, self._icustay_fname), header=True)
            patient = patient.select(*[self.patient_key, self.icustay_key, self._hadm_key])  # icuunit intime
            multi_hosp = (
                patient.groupBy(self.patient_key)
                .agg(F.count(self._hadm_key).alias("count"))
                .filter(F.col("count") > 1)
                .select(self.patient_key)
                .rdd.flatMap(lambda row: row)
                .collect()
            )
            # multiple hosp

            dialysis_tables = self.task_itemids["dialysis"]["fname"]  # Only treatment for dialysis
            dialysis_code = self.task_itemids["dialysis"]["code"][0]
            excludes = self.task_itemids["dialysis"]["exclude"]

            io = spark.read.csv(os.path.join(self.data_dir, dialysis_tables), header=True)
            io = io.drop(*excludes)

            io_dialysis = io.filter(F.col(dialysis_code) != 0)
            io_dialysis = io_dialysis.join(patient, on=self.icustay_key, how="left")

            dialysis_multihosp = (
                io_dialysis.filter(F.col(self.patient_key).isin(multi_hosp))
                .select(self.patient_key)
                .rdd.flatMap(lambda row: row)
                .collect()
            )

            io_dialysis = io_dialysis.drop(self.patient_key)

            def dialysis_time(table, timecolumn):
                return table.withColumn("_DIALYSIS_TIME", F.col(timecolumn)).select(
                    self.icustay_key, "_DIALYSIS_TIME"
                )

            io_dialysis = dialysis_time(io_dialysis, self.task_itemids["dialysis"]["timestamp"])
            io_dialysis = io_dialysis.groupBy(self.icustay_key).agg(
                F.min("_DIALYSIS_TIME").alias("_DIALYSIS_TIME")
            )
            io_dialysis = io_dialysis.select([self.icustay_key, "_DIALYSIS_TIME"])
            merge = merge.join(io_dialysis, on=self.icustay_key, how="left")
            merge = merge.filter(F.isnull("_DIALYSIS_TIME") | (F.col("_DIALYSIS_TIME") > F.col(timestamp)))
            merge = merge.drop("_DIALYSIS_TIME")

        # For Creatinine task, eliminate icus if patient went through dialysis treatment
        # before (obs_size + pred_size) timestamp

        # Cohort with events within (obs_size + gap_size) - (obs_size + pred_size)
        merge = merge.filter(((self.obs_size + self.gap_size) * 60) <= F.col(timestamp)).filter(
            ((self.obs_size + self.pred_size) * 60) >= F.col(timestamp)
        )

        # Average value of events
        value_agg = merge.groupBy(self.icustay_key).agg(
            F.mean(value).alias("avg_value")
        )  # TODO: mean/min/max?

        # Labeling
        if task == "bilirubin":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 1.2, 0)
                .when((value_agg.avg_value >= 1.2) & (value_agg.avg_value < 2.0), 1)
                .when((value_agg.avg_value >= 2.0) & (value_agg.avg_value < 6.0), 2)
                .when((value_agg.avg_value >= 6.0) & (value_agg.avg_value < 12.0), 3)
                .when(value_agg.avg_value >= 12.0, 4),
            )
        elif task == "platelets":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value >= 150, 0)
                .when((value_agg.avg_value >= 100) & (value_agg.avg_value < 150), 1)
                .when((value_agg.avg_value >= 50) & (value_agg.avg_value < 100), 2)
                .when((value_agg.avg_value >= 20) & (value_agg.avg_value < 50), 3)
                .when(value_agg.avg_value < 20, 4),
            )

        elif task == "creatinine":
            value_agg = value_agg.join(
                patient.select([self.patient_key, self.icustay_key]), on=self.icustay_key, how="left"
            )
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 1.2, 0)
                .when((value_agg.avg_value >= 1.2) & (value_agg.avg_value < 2.0), 1)
                .when((value_agg.avg_value >= 2.0) & (value_agg.avg_value < 3.5), 2)
                .when((value_agg.avg_value >= 3.5) & (value_agg.avg_value < 5), 3)
                .when(value_agg.avg_value >= 5, 4),
            )
            value_agg = value_agg.filter(~F.col(self.patient_key).isin(dialysis_multihosp))
            value_agg = value_agg.drop(self.patient_key)

        elif task == "wbc":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 4, 0)
                .when((value_agg.avg_value >= 4) & (value_agg.avg_value <= 12), 1)
                .when((value_agg.avg_value > 12), 2),
            )

        cohorts = cohorts.join(value_agg.select(self.icustay_key, task), on=self.icustay_key, how="left")

        return cohorts

    def infer_data_extension(self) -> str:
        if len(glob.glob(os.path.join(self.data_dir, "*.csv.gz"))) == 31:
            ext = ".csv.gz"
        elif len(glob.glob(os.path.join(self.data_dir, "*.csv"))) == 31:
            ext = ".csv"
        else:
            raise AssertionError(
                "Provided data directory is not correct. Please check if --data is correct. "
                "--data: {}".format(self.data_dir)
            )

        logger.info("Data extension is set to '{}'".format(ext))

        return ext



import argparse
import logging
import os
import sys

from ehrs import EHR_REGISTRY
from pyspark.sql import SparkSession

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    stream=sys.stdout,
)
logger = logging.getLogger(__name__)


def get_parser():
    parser = argparse.ArgumentParser()
    # task
    parser.add_argument(
        "--task",
        type=str,
        default=None,
        help="specific task from the all_features pipeline." "if not set, run the all_features steps.",
    )
    parser.add_argument("--dest", default="outputs", type=str, metavar="DIR", help="output directory")

    # data
    parser.add_argument(
        "--ehr",
        type=str,
        required=True,
        choices=["mimiciii", "mimiciv", "eicu"],
        help="name of the ehr system to be processed.",
    )
    parser.add_argument(
        "--data",
        metavar="DIR",
        default=None,
        help="directory containing data files of the given ehr (--ehr)."
        "if not given, try to download from the internet.",
    )
    parser.add_argument(
        "--ext",
        type=str,
        default=None,
        help="extension for ehr data to look for. " "if not given, try to infer from --data",
    )
    parser.add_argument(
        "--ccs",
        type=str,
        default=None,
        help="path to `ccs_multi_dx_tool_2015.csv`" "if not given, try to download from the internet.",
    )
    parser.add_argument(
        "--gem",
        type=str,
        default=None,
        help="path to `icd10cmtoicd9gem.csv`" "if not given, try to download from the internet.",
    )

    parser.add_argument(
        "-c", "--cache", action="store_true", help="whether to load data from cache if exists"
    )

    # misc
    parser.add_argument("--max_event_size", type=int, default=256, help="max event size to crop to")
    parser.add_argument(
        "--min_event_size",
        type=int,
        default=5,
        help="min event size to skip small samples",
    )
    parser.add_argument("--min_age", type=int, default=18, help="min age to skip too young patients")
    parser.add_argument("--max_age", type=int, default=None, help="max age to skip too old patients")
    parser.add_argument("--obs_size", type=int, default=12, help="observation window size by the hour")
    parser.add_argument("--gap_size", type=int, default=12, help="time gap window size by the hour")
    parser.add_argument("--pred_size", type=int, default=24, help="prediction window size by the hour")
    parser.add_argument(
        "--long_term_pred_size",
        type=int,
        default=336,
        help="prediction window size by the hour (for long term mortality task)",
    )
    parser.add_argument(
        "--first_icu",
        action="store_true",
        help="whether to use only the first icu or not",
    )

    # tasks
    parser.add_argument("--mortality", action="store_true", help="whether to include mortality task or not")
    parser.add_argument(
        "--long_term_mortality",
        action="store_true",
        help="whether to include long term mortality task or not",
    )
    parser.add_argument("--los_3day", action="store_true", help="whether to include 3-day los task or not")
    parser.add_argument("--los_7day", action="store_true", help="whether to include 7-day los task or not")
    parser.add_argument(
        "--readmission", action="store_true", help="whether to include readmission task or not"
    )
    parser.add_argument(
        "--final_acuity", action="store_true", help="whether to include final acuity task or not"
    )
    parser.add_argument(
        "--imminent_discharge", action="store_true", help="whether to include imminent discharge task or not"
    )
    parser.add_argument("--diagnosis", action="store_true", help="whether to include diagnosis task or not")
    parser.add_argument("--creatinine", action="store_true", help="whether to include creatinine task or not")
    parser.add_argument("--bilirubin", action="store_true", help="whether to include bilirubin task or not")
    parser.add_argument("--platelets", action="store_true", help="whether to include platelets task or not")
    parser.add_argument(
        "--wbc", action="store_true", help="whether to include blood white blood cell count task or not"
    )

    parser.add_argument(
        "--chunk_size",
        type=int,
        default=1024,
        help="chunk size to read large csv files",
    )
    parser.add_argument("--bins", type=int, default=20, help="num buckets to bin time intervals by")

    parser.add_argument(
        "--max_event_token_len", type=int, default=192, help="max token length for each event (Hierarchical)"
    )

    parser.add_argument(
        "--max_patient_token_len", type=int, default=8192, help="max token length for each patient (Flatten)"
    )

    parser.add_argument("--num_threads", type=int, default=8, help="number of threads to use")

    # CodeEmb / Feature select
    parser.add_argument(
        "--emb_type",
        type=str,
        choices=["codebase", "textbase"],
        default="textbase",
        help="feature embedding type, codebase model = [SAND, Rajikomar], textbase model = [GenHPF, DescEmb]",
    )
    parser.add_argument(
        "--feature",
        choices=["select", "all_features"],
        default="all_features",
        help="pre-define feature select or all_features table use",
    )
    parser.add_argument("--bucket_num", type=int, default=10, help="feature bucket num")
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    if not os.path.exists(args.dest):
        os.makedirs(args.dest)

    ehr = EHR_REGISTRY[args.ehr](args)

    spark = (
        SparkSession.builder.master(f"local[{args.num_threads}]")
        .config("spark.driver.memory", "100g")
        .config("spark.driver.maxResultSize", "10g")
        .config("spark.network.timeout", "100s")
        .appName("Main_Preprocess")
        .getOrCreate()
    )
    ehr.run_pipeline(spark)


if __name__ == "__main__":
    main()



import torch

class GradMultiply(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale):
        ctx.scale = scale
        res = x.new(x)
        return res
    
    @staticmethod
    def backward(ctx, grad):
        return grad * ctx.scale, None


import logging

from .configs import (
    BaseConfig,
    Config,
    CommonConfig,
    DistributedTrainingConfig,
    DatasetConfig,
    CheckpointConfig,
)
from .constants import ChoiceEnum

logger = logging.getLogger(__name__)

__all__ = [
    "BaseConfig",
    "Config",
    "CommonConfig",
    "DistributedTrainingConfig",
    "DatasetConfig",
    "CheckpointConfig",
    "ChoiceEnum",
]


import numpy as np
import pandas as pd


def col_name_add(x, cate_col):
    if not (x == "nan" or x == pd.isnull(x)):
        return cate_col + "_" + str(x)
    else:
        return x  # return nan


def q_cut(x, cuts):
    unique_var = len(np.unique([i for i in x]))
    nunique = len(pd.qcut(x, min(unique_var, cuts), duplicates="drop").cat.categories)
    output = pd.qcut(x, min(unique_var, cuts), labels=range(1, min(nunique, cuts) + 1), duplicates="drop")
    return output



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

import torch
from torch.nn.modules.loss import _Loss

from genhpf.configs import BaseConfig
from genhpf.models.genhpf import GenHPF


@dataclass
class CriterionConfig(BaseConfig):
    task_names: Optional[List[str]] = field(
        default=None, metadata={"help": "a list of task names for multi-task learning"}
    )
    num_labels: Optional[List[int]] = field(
        default=None, metadata={"help": "a list of number of labels for each task"}
    )


class BaseCriterion(_Loss):
    def __init__(self, cfg: CriterionConfig):
        super().__init__()
        self.cfg = cfg

        self.task_names = cfg.task_names
        self.num_labels = cfg.num_labels

    @classmethod
    def build_criterion(cls, cfg: CriterionConfig):
        """Construct a new criterion instance."""
        return cls(cfg)

    def compute_loss(
        self, logits: torch.Tensor, targets: torch.Tensor, sample=None, net_output=None, model=None
    ) -> Tuple[torch.Tensor, List[float]]:
        """Compute the loss given the logits and targets from the model."""
        raise NotImplementedError("Criterion must implement the `compute_loss` method")

    def get_sample_size(self, sample, targets: torch.Tensor) -> int:
        """Get the sample size, which is used as the denominator for the gradient."""
        raise NotImplementedError("Criterion must implement the `get_sample_size` method")

    def get_logging_outputs(
        self, logging_output, logits: torch.Tensor, targets: torch.Tensor, sample=None
    ) -> List[Dict[str, Any]]:
        """
        Get the logging output to display while training
        """
        raise NotImplementedError("Criterion must implement the `get_logging_outputs` method")

    def forward(self, model: GenHPF, sample, return_net_output=False):
        """Compute the loss for the given sample.

        Returns a tuple with three elements:
        1. the loss
        2. the sample size, which is used as the denominator for the gradient
        3. logging outputs to display while training
        """
        net_output = model(**sample["net_input"])
        logits = model.get_logits(sample, net_output)
        targets = model.get_targets(sample, net_output)

        loss, losses_to_log = self.compute_loss(
            logits, targets, sample=sample, net_output=net_output, model=model
        )
        sample_size = self.get_sample_size(sample, targets)

        logging_output = {}
        if len(losses_to_log) > 1:
            logging_output["loss"] = loss.item()
            for i, l in enumerate(losses_to_log):
                logging_output[f"loss_{i}"] = l
        else:
            logging_output["loss"] = losses_to_log[0]
        logging_output["sample_size"] = sample_size
        logging_output = self.get_logging_outputs(logging_output, logits, targets, sample)

        if return_net_output:
            return loss, sample_size, logging_output, net_output
        else:
            return loss, sample_size, logging_output

    @staticmethod
    def reduce_metrics(stats: Dict[str, Any], prefix: str = None) -> None:
        """Aggregate logging outputs from data parallel training."""
        raise NotImplementedError



import contextlib
import uuid
from collections import defaultdict, OrderedDict
from typing import Callable, List, Optional

from .meters import *

_aggregators = OrderedDict()
_active_aggregators = OrderedDict()
_active_aggregators_cnt = defaultdict(lambda: 0)

def reset() -> None:
    _aggregators.clear()
    _active_aggregators.clear()
    _active_aggregators_cnt.clear()

    _aggregators['default'] = MetersDict()
    _active_aggregators['default'] = _aggregators['default']
    _active_aggregators_cnt['default'] = 1

reset()

@contextlib.contextmanager
def aggregate(name=None, new_root=False):
    if name is None:
        name = str(uuid.uuid4())
        assert name not in _aggregators
        agg = MetersDict()
    else:
        assert name != 'default'
        agg = _aggregators.setdefault(name, MetersDict())
    
    if new_root:
        backup_aggregators = _active_aggregators.copy()
        _active_aggregators.clear()
        backup_aggregators_cnt = _active_aggregators_cnt.copy()
        _active_aggregators_cnt.clear()
    
    _active_aggregators[name] = agg
    _active_aggregators_cnt[name] += 1

    yield agg

    _active_aggregators_cnt[name] -= 1
    if _active_aggregators_cnt[name] == 0 and name in _active_aggregators:
        del _active_aggregators[name]
    
    if new_root:
        _active_aggregators.clear()
        _active_aggregators.update(backup_aggregators)
        _active_aggregators_cnt.clear()
        _active_aggregators_cnt.update(backup_aggregators_cnt)

def get_active_aggregators() -> List[MetersDict]:
    return list(_active_aggregators.values())

def log_scalar(
    key,
    value,
    weight=1,
    priority=10,
    round=None
):
    for agg in get_active_aggregators():
        if key not in agg:
            agg.add_meter(key, AverageMeter(round=round), priority)
        agg[key].update(value, weight)

def log_scalar_sum(
    key,
    value,
    priority=10,
    round=None
):
    for agg in get_active_aggregators():
        if key not in agg:
            agg.add_meter(key, SumMeter(round=round), priority)
        agg[key].update(value)

def log_derived(
    key,
    fn: Callable[[MetersDict], float],
    priority=20
):
    for agg in get_active_aggregators():
        if key not in agg:
            agg.add_meter(key, MetersDict._DerivedMeter(fn), priority)

def log_speed(
    key,
    value,
    priority=30,
    round=None
):
    for agg in get_active_aggregators():
        if key not in agg:
            agg.add_meter(key, TimeMeter(round=round), priority)
            agg[key].reset()
        else:
            agg[key].update(value)

def log_start_time(
    key,
    priority=40,
    round=None
):
    for agg in get_active_aggregators():
        if key not in agg:
            agg.add_meter(key, StopwatchMeter(round=round), priority)
    agg[key].start()

def log_stop_time(
    key,
    weight=0.0,
    prehook=None
):
    for agg in get_active_aggregators():
        if key in agg:
            agg[key].stop(weight, prehook)

def log_custom(
    new_meter_fn: Callable[[], Meter],
    key,
    *args,
    priority=50,
    **kwargs,
):
    for agg in get_active_aggregators():
        if key not in agg:
            agg.add_meter(key, new_meter_fn(), priority)
        agg[key].update(*args, **kwargs)

def reset_meter(name, key) -> None:
    meter = get_meter(name, key)
    if meter is not None:
        meter.reset()

def reset_meters(name) -> None:
    meters = get_meters(name)
    if meters is not None:
        meters.reset()

def get_meter(name, key) -> Meter:
    if name not in _aggregators:
        return None
    return _aggregators[name].get(key, None)

def get_meters(name) -> MetersDict:
    return _aggregators.get(name, None)

def get_smoothed_values(name) -> Dict[str, float]:
    return _aggregators[name].get_smoothed_values()

def state_dict():
    return OrderedDict([(name, agg.state_dict()) for name, agg in _aggregators.items()])


import logging
import os
import pprint
import sys
from itertools import chain

import polars as pl
import torch.distributed
import torch.utils.data

logging.basicConfig(
    format="%(asctime)s | %(levelname)s %(name)s %(message)s)))",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    stream=sys.stdout,
)
logger = logging.getLogger("genhpf.test")

import hydra
import torch
from hydra.core.hydra_config import HydraConfig
from omegaconf import OmegaConf, open_dict

from genhpf import criterions, models
from genhpf.configs import Config
from genhpf.configs.initialize import add_defaults, hydra_init
from genhpf.datasets import load_dataset
from genhpf.loggings import metrics, progress_bar
from genhpf.utils import distributed_utils, utils


def main(cfg: Config) -> None:
    assert (
        cfg.checkpoint.load_checkpoint is not None
    ), "Please specify the checkpoint to load with `checkpoint.load_checkpoint`"

    assert cfg.dataset.batch_size is not None, "batch_size must be specified"
    metrics.reset()

    use_cuda = torch.cuda.is_available()

    if use_cuda:
        torch.cuda.set_device(cfg.distributed_training.device_id)

    if cfg.distributed_training.distributed_world_size > 1:
        data_parallel_world_size = distributed_utils.get_data_parallel_world_size()
    else:
        data_parallel_world_size = 1

    # print args
    logger.info(pprint.pformat(dict(cfg)))

    # load model
    model = models.build_model(cfg.model)
    logger.info(f"loading model from {cfg.checkpoint.load_checkpoint}")
    model_state_dict = torch.load(cfg.checkpoint.load_checkpoint, map_location="cpu")["model"]
    model.load_state_dict(model_state_dict, strict=True)
    logger.info(f"loaded model from {cfg.checkpoint.load_checkpoint}")

    logger.info(model)
    logger.info(f"model: {model.__class__.__name__}")
    logger.info(
        "num. shared model params: {:,} (num. trained: {:,})".format(
            sum(p.numel() for p in model.parameters()),
            sum(p.numel() for p in model.parameters() if p.requires_grad),
        )
    )

    # Move model to GPU
    model.eval()
    if use_cuda:
        model.cuda()

    # build criterion
    criterion = criterions.build_criterion(cfg.criterion)
    criterion.eval()

    def _fp_convert_sample(sample):
        def apply_float(t):
            if t.dtype in [torch.float64, torch.float32, torch.int16]:
                return t.to(dtype=torch.float)
            return t

        sample = utils.apply_to_sample(apply_float, sample)

        return sample

    assert cfg.dataset.test_subset is not None, "Please specify the test subset with `dataset.test_subset`"
    test_subsets = cfg.dataset.test_subset.split(",")

    if cfg.dataset.combine_test_subsets:
        datasets = [("combined-test", load_dataset(cfg.dataset.data, test_subsets, cfg))]
    else:
        datasets = [
            (subset.strip(), load_dataset(cfg.dataset.data, [subset], cfg)) for subset in test_subsets
        ]

    output_meds_predictions = False
    if cfg.dataset.data_format == "meds" and cfg.meds.output_predictions:
        if len(test_subsets) > 1:
            raise NotImplementedError(
                "MEDS dataset does not currently support multiple test subsets when `output_predictions` "
                "is enabled. Please specify only one test subset."
            )
        if cfg.dataset.combine_test_subsets:
            raise NotImplementedError(
                "MEDS dataset does not currently support `combine_test_subsets` when `output_predictions` "
                "is enabled. Please set `dataset.combine_test_subsets` to False."
            )
        if len(cfg.criterion.task_names) > 1:
            raise NotImplementedError(
                "MEDS dataset does not currently support multiple tasks when `output_predictions` "
                "is enabled. Please specify only one task."
            )
        if cfg.criterion.num_labels[0] > 1:
            raise NotImplementedError(
                "MEDS dataset currently only supports binary classification when `output_predictions` "
                "is enabled. Please specify only one label by setting `criterion.num_labels` to 1."
            )

        assert (
            cfg.meds.labels_dir is not None and cfg.meds.output_dir is not None
        ), "Please specify labels_dir and output_dir in the MEDS config to output predictions."
        assert data_parallel_world_size == 1, (
            "MEDS dataset does not currently support distributed testing when `output_predictions` "
            "is enabled. Please set `distributed_training.distributed_world_size` to 1."
        )

        output_meds_predictions = True

        labels = pl.read_parquet(os.path.join(cfg.meds.labels_dir, f"{test_subsets[0]}/*.parquet"))
        labels = labels.sort(by=["subject_id", "prediction_time"])
        labels = labels.with_columns(pl.col("subject_id").cum_count().over("subject_id").alias("suffix"))
        labels = labels.with_columns(
            pl.col("subject_id").cast(str) + "_" + pl.col("suffix").cast(str).alias("subject_id")
        )
        labels = labels.drop("suffix")
        labels = labels.select(["subject_id", "prediction_time", "boolean_value"])

        meds_pred_output = {
            "subject_id": [],
            "predicted_boolean_value": [],
            "predicted_boolean_probability": [],
        }

    for subset, dataset in datasets:
        logger.info(f"begin validation on '{subset}' subset")

        # initialize data iterator
        batch_sampler = (
            torch.utils.data.DistributedSampler(dataset, shuffle=False)
            if torch.distributed.is_initialized()
            else None
        )
        batch_iterator = torch.utils.data.DataLoader(
            dataset,
            batch_size=cfg.dataset.batch_size,
            shuffle=False,
            num_workers=cfg.dataset.num_workers,
            collate_fn=dataset.collator,
            sampler=batch_sampler,
        )

        progress = progress_bar.progress_bar(
            batch_iterator,
            log_format=cfg.common.log_format,
            log_interval=cfg.common.log_interval,
            log_file=cfg.common.log_file,
            epoch=0,
            default_log_format=("tqdm" if cfg.common.no_progress_bar else "simple"),
            wandb_project=(
                cfg.common.wandb_project if distributed_utils.is_master(cfg.distributed_training) else None
            ),
            wandb_entity=(
                cfg.common.wandb_entity if distributed_utils.is_master(cfg.distributed_training) else None
            ),
            wandb_run_name=os.environ.get("WANDB_NAME", os.path.basename(cfg.checkpoint.save_dir)),
        )

        log_outputs = []
        for i, sample in enumerate(progress):
            with torch.no_grad():
                sample = utils.prepare_sample(sample)
                sample = _fp_convert_sample(sample)
                _loss, _sample_size, log_output, net_output = criterion(model, sample, return_net_output=True)
                log_outputs.append(log_output)
                if output_meds_predictions:
                    meds_pred_output["subject_id"].extend(sample["id"])
                    logits = model.get_logits(sample, net_output)
                    probs = torch.sigmoid(logits).view(-1).cpu()
                    meds_pred_output["predicted_boolean_probability"].extend(probs.tolist())
                    meds_pred_output["predicted_boolean_value"].extend(
                        (probs >= cfg.criterion.threshold).int().tolist()
                    )

        if output_meds_predictions:
            meds_pred_output = pl.DataFrame(meds_pred_output)
            meds_pred_output = meds_pred_output.join(labels, on="subject_id", how="left")
            meds_pred_output = meds_pred_output.select(
                [
                    pl.col("subject_id"),
                    pl.col("prediction_time"),
                    pl.col("boolean_value"),
                    pl.col("predicted_boolean_value"),
                    pl.col("predicted_boolean_probability"),
                ]
            )
            meds_pred_output = meds_pred_output.with_columns(
                pl.col("subject_id").map_elements(lambda x: x.split("_")[0], return_dtype=pl.String).cast(int)
            )
            meds_pred_output = (
                meds_pred_output.with_columns(pl.col("predicted_boolean_value").cast(bool))
            )
            if not os.path.exists(cfg.meds.output_dir):
                os.makedirs(cfg.meds.output_dir)
            meds_pred_output.write_parquet(os.path.join(cfg.meds.output_dir, f"{subset}.parquet"))

        if data_parallel_world_size > 1:
            log_outputs = distributed_utils.all_gather_list(
                log_outputs,
                max_size=cfg.common.all_gather_list_size,
                group=distributed_utils.get_data_parallel_group(),
            )
            log_outputs = list(chain.from_iterable(log_outputs))

        with metrics.aggregate(new_root=True) as agg:
            criterion.__class__.reduce_metrics(log_outputs)
            del log_outputs
            log_outputs = agg.get_smoothed_values()

        if hasattr(criterion, "post_validate"):
            stats = criterion.post_validate(stats=log_outputs, agg=agg)

        progress.print(stats, tag=subset, step=None)


@hydra.main(config_path=os.path.join("..", "configs"), config_name="config")
def hydra_main(cfg: Config) -> None:
    add_defaults(cfg)

    with open_dict(cfg):
        # make hydra logging work with ddp (see # see https://github.com/facebookresearch/hydra/issues/1126)
        cfg.job_logging_cfg = OmegaConf.to_container(HydraConfig.get().job_logging, resolve=True)

    cfg = OmegaConf.create(OmegaConf.to_container(cfg, resolve=True, enum_to_str=True))
    OmegaConf.set_struct(cfg, True)

    distributed_utils.call_main(cfg, main)


def cli_main():
    try:
        from hydra._internal.utils import get_args

        cfg_name = get_args().config_name or "config"
    except Exception:
        logger.warning("Failed to get config name from hydra args")
        cfg_name = "config"
    hydra_init(cfg_name)
    hydra_main()


if __name__ == "__main__":
    cli_main()



import math
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn.functional as F

import genhpf.utils.utils as utils
from genhpf.criterions import BaseCriterion, register_criterion
from genhpf.criterions.criterion import CriterionConfig
from genhpf.loggings import metrics
from genhpf.loggings.meters import safe_round


@dataclass
class Wav2Vec2CriterionConfig(CriterionConfig):
    loss_weights: Optional[List[float]] = field(
        default=None, metadata={"help": "weights for additional loss terms (not first one)"}
    )


@register_criterion("wav2vec2_criterion", dataclass=Wav2Vec2CriterionConfig)
class Wav2Vec2Criterion(BaseCriterion):
    def __init__(self, cfg: Wav2Vec2CriterionConfig):
        super().__init__(cfg)

        self.loss_weights = cfg.loss_weights

    def compute_loss(
        self, logits: torch.Tensor, targets: torch.Tensor, sample=None, net_output=None, model=None
    ) -> Tuple[torch.Tensor, List[float]]:
        """Compute the loss given the logits and targets from the model."""

        losses = []

        loss = F.cross_entropy(logits, targets, reduction="sum")
        losses.append(loss.detach().item())

        sample_size = self.get_sample_size(sample, targets)

        if self.loss_weights is not None:
            assert hasattr(model, "get_extra_losses")
            extra_losses = model.get_extra_losses(net_output)
            if torch.is_tensor(extra_losses):
                extra_losses = [extra_losses]
            if len(self.loss_weights) == 1 and len(extra_losses) != 1:
                self.loss_weights = [self.weights[0]] * len(extra_losses)
            assert len(extra_losses) == len(
                self.loss_weights
            ), f"{len(extra_losses)}, {len(self.loss_weights)}"
            for p, coef in zip(extra_losses, self.loss_weights):
                if coef != 0 and p is not None:
                    p = coef * p.float() * sample_size
                    loss += p
                    losses.append(p.detach().item())

        return loss, losses

    def get_sample_size(self, sample, targets):
        """
        Get the sample size, which is used as the denominator for the gradient
        """
        if "sample_size" in sample:
            sample_size = sample["sample_size"]
        else:
            sample_size = targets.numel()
        return sample_size

    def get_logging_outputs(
        self, logging_output, logits: torch.Tensor, targets: torch.Tensor, sample=None
    ) -> List[Dict[str, Any]]:
        """
        Get the logging output to display while training
        """
        with torch.no_grad():
            if logits.numel() == 0:
                corr = 0
                count = 0
            else:
                assert logits.dim() > 1, logits.shape
                max = logits.argmax(-1) == 0
                min = logits.argmin(-1) == 0

                both = max & min
                corr = max.long().sum().item() - both.long().sum().item()
                count = float(max.numel())

            logging_output["correct"] = corr
            logging_output["count"] = count
        return logging_output

    @staticmethod
    def reduce_metrics(logging_outputs: List[Dict[str, Any]], prefix: str = None) -> None:
        """Aggregate logging outputs from data parallel training."""
        if prefix is None:
            prefix = ""
        elif prefix is not None and not prefix.endswith("_"):
            prefix = prefix + "_"

        loss_sum = utils.item(sum(log.get("loss", 0) for log in logging_outputs))

        sample_size = utils.item(sum(log.get("sample_size", 0) for log in logging_outputs))

        metrics.log_scalar(f"{prefix}loss", loss_sum / (sample_size or 1) / math.log(2), sample_size, round=3)

        correct = sum(log.get("correct", 0) for log in logging_outputs)
        metrics.log_scalar(f"_{prefix}correct", correct)

        total = sum(log.get("count", 0) for log in logging_outputs)
        metrics.log_scalar(f"_{prefix}total", total)

        if total > 0:
            metrics.log_derived(
                f"{prefix}accuracy",
                lambda meters: safe_round(meters[f"_{prefix}correct"].sum / meters[f"_{prefix}total"].sum, 5)
                if meters[f"_{prefix}total"].sum > 0
                else float("nan"),
            )

        builtin_keys = {"loss", "sample_size", "correct", "count"}

        for k in logging_outputs[0]:
            if k not in builtin_keys:
                val = sum(log.get(k, 0) for log in logging_outputs)
                if k.startswith("loss"):
                    metrics.log_scalar(
                        prefix + k, val / (sample_size or 1) / math.log(2), sample_size, round=3
                    )
                else:
                    metrics.log_scalar(prefix + k, val / len(logging_outputs), round=3)



import logging
from dataclasses import dataclass
from typing import List

import torch
import torch.nn as nn
from omegaconf import II

from genhpf.models import register_model
from genhpf.models.genhpf import GenHPF, GenHPFConfig

logger = logging.getLogger(__name__)


@dataclass
class GenHPFPredictorConfig(GenHPFConfig):
    tasks: List[str] = II("criterion.task_names")
    num_labels: List[int] = II("criterion.num_labels")


@register_model("genhpf_predictor", dataclass=GenHPFPredictorConfig)
class GenHPFPredictor(GenHPF):
    def __init__(self, cfg: GenHPFPredictorConfig):
        super().__init__(cfg)

        self.num_labels = cfg.num_labels
        self.tasks = cfg.tasks
        assert len(self.num_labels) == len(
            cfg.tasks
        ), "The number of num_labels must be equal to the number of tasks"

        self.final_proj = nn.ModuleDict()
        for i, task in enumerate(cfg.tasks):
            self.final_proj[task] = nn.Linear(cfg.agg_embed_dim, self.num_labels[i])

    def get_logits(self, sample, net_output):
        if len(self.tasks) == 1:
            return net_output[self.tasks[0]]
        else:
            return net_output

    def get_targets(self, sample, net_output):
        if len(self.tasks) == 1:
            return sample["label"][self.tasks[0]]
        else:
            return sample["label"]

    def forward(
        self,
        input_ids: torch.Tensor,
        type_ids: torch.Tensor = None,
        dpe_ids: torch.Tensor = None,
        padding_mask: torch.Tensor = None,
        **kwargs,
    ):
        x, padding_mask = super().forward(
            input_ids=input_ids, type_ids=type_ids, dpe_ids=dpe_ids, padding_mask=padding_mask, **kwargs
        )

        if padding_mask is not None and padding_mask.any():
            x[padding_mask] = 0
        x = torch.div(x.sum(dim=1), (x != 0).sum(dim=1))

        ret = {}
        for task, proj in self.final_proj.items():
            ret[task] = proj(x)

        return ret

    def get_finetuning_parameter_names(self):
        ret = []
        ret.extend(["final_proj" + "." + x[0] for x in self.final_proj.named_parameters()])
        return ret



import importlib
import os

from .ehr import EHR

EHR_REGISTRY = {}

__all__ = "EHR"


def register_ehr(name):
    def register_ehr(cls):
        if name in EHR_REGISTRY:
            raise ValueError("Cannot register duplicate EHR ({})".format(name))
        EHR_REGISTRY[name] = cls

        return cls

    return register_ehr


def import_ehrs(ehrs_dir, namespace):
    for file in os.listdir(ehrs_dir):
        path = os.path.join(ehrs_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            ehrs_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + ehrs_name)


# automatically import any Python files in the ehrs/ directory
ehrs_dir = os.path.dirname(__file__)
import_ehrs(ehrs_dir, "ehrs")



import argparse
import random
import os
import logging
import h5pickle
from tqdm import tqdm

logger = logging.getLogger(__name__)

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "data", type=str, help="path to the .h5 file containing the data"
    )
    parser.add_argument(
        "label", type=str, help="path to the .csv file containing the labels corresponded to the data"
    )
    parser.add_argument(
        "--dest", default=".", type=str, metavar="DIR", help="output directory"
    )
    parser.add_argument(
        "--prefix", default="", type=str, help="prefix for the output files"
    )
    parser.add_argument(
        "--valid-percent",
        default=0.1,
        type=float,
        metavar="D",
        help="percentage of the data to use as validation and test set (between 0 and 0.5)"
    )
    parser.add_argument("--seed", default=42, type=int, metavar="N", help="random seed")
    return parser

def main(args):
    assert 0 <= args.valid_percent <= 0.5

    data_path = os.path.realpath(args.data)
    label_path = os.path.realpath(args.label)
    prefix = args.prefix
    if len(prefix) > 0 and not prefix.endswith("_"):
        prefix += "_"
    rand = random.Random(args.seed)

    if not os.path.exists(args.dest):
        os.makedirs(args.dest)
    
    with (
        open(os.path.join(args.dest, f"{prefix}train.tsv"), "w") as train_f,
        open(os.path.join(args.dest, f"{prefix}valid.tsv"), "w") as valid_f,
        open(os.path.join(args.dest, f"{prefix}test.tsv"), "w") as test_f
    ):
        print(data_path, file=train_f)
        print(label_path, file=train_f)
        print(data_path, file=valid_f)
        print(label_path, file=valid_f)
        print(data_path, file=test_f)
        print(label_path, file=test_f)

        def write(subject_ids, dest, split):
            for subject_id in tqdm(subject_ids, total=len(subject_ids), desc=split):
                print(subject_id, file=dest)

        data = h5pickle.File(data_path, "r")["ehr"]
        subject_ids = list(data.keys())

        rand.shuffle(subject_ids)

        valid_len = int(len(subject_ids) * args.valid_percent)
        test_len = int(len(subject_ids) * args.valid_percent)
        train_len = len(subject_ids) - valid_len - test_len

        train = subject_ids[:train_len]
        valid = subject_ids[train_len: train_len + valid_len]
        test = subject_ids[train_len + valid_len:]

        write(train, train_f, split=f"{prefix}train")
        write(valid, valid_f, split=f"{prefix}valid")
        write(test, test_f, split=f"{prefix}test")

if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(args)